from argparse import ArgumentParser
from calendar import month_name
from datetime import datetime
from os import makedirs
from os.path import abspath, split

import numpy as np

from rad_tools import __version__ as version
from rad_tools.io.tb2j import read_exchange_model
from rad_tools.routines import OK, RESET, YELLOW


def manager(
    output_name="template",
    input_filename=None,
    R_vector=None,
    max_distance=None,
    min_distance=None,
    distance=None,
    verbose=False,
):
    r"""
    Main function call of the rad-make-template.py script.

    Full documentation on the behaviour is available
    :ref:`here <rad-make-template>`. Parameters of the function directly
    corresponds to the arguments of the script.

    If you want to have the behaviour of the rad-make-template.py script
    but in a format of a function call use this function.
    """

    if distance is not None:
        min_distance = distance
        max_distance = distance

    if R_vector is not None:
        R_vector = np.array(R_vector[: len(R_vector) // 3 * 3], dtype=int).reshape(
            (len(R_vector) // 3, 3)
        )
        R_vector = list(map(tuple, R_vector.tolist()))

    try:
        makedirs(split(output_name)[0])
    except FileExistsError:
        pass
    except FileNotFoundError:
        pass

    template = (
        "=" * 20
        + "\n"
        + "Neighbors template:\n"
        + "i j R_a R_b R_c\n"
        + "-" * 20
        + "\n"
        + "J1 $J_1$\n"
        + "atom1 atom2  0  0  0\n"
        + "atom1 atom2  1  0  0\n"
        + "atom1 atom1 -1  0  2\n"
        + "-" * 20
        + "\n"
        + "J2\n"
        + "atom2 atom1  9  5 -3\n"
        + "atom1 atom2  1  4  0\n"
        + "atom2 atom2  1  0  2\n"
        + "=" * 20
        + "\n"
    )
    cd = datetime.now()
    with open(f"{output_name}.txt", "w") as file:
        if input_filename is None:
            file.write(
                f"Template is created "
                + f"on {cd.day} {month_name[cd.month]} {cd.year}"
                + f" at {cd.hour}:{cd.minute}:{cd.second} by rad-tools {version}\n\n"
            )

            file.write(template)
        else:
            file.write(
                f"Template is created based on the file: {input_filename}\n"
                + f"on {cd.day} {month_name[cd.month]} {cd.year}"
                + f" at {cd.hour}:{cd.minute}:{cd.second} by rad-tools {version}\n\n"
            )

            model = read_exchange_model(input_filename, quiet=not verbose)
            model.filter(
                min_distance=min_distance, max_distance=max_distance, R_vector=R_vector
            )
            file.write(
                "=" * 20
                + "\n"
                + "Neighbors template:\n"
                + "i j R_a R_b R_c\n"
                + "-" * 20
                + "\n"
                + "Name placeholder"
                + "\n"
            )
            for atom1, atom2, R in model.bonds:
                file.write(
                    f"{atom1:4} {atom2:4} " + f"{R[0]:3.0f} {R[1]:3.0f} {R[2]:3.0f}\n"
                )
            file.write("=" * 20 + "\n")
    print(f"{OK}Template draft is in " + f"{abspath(output_name)}.txt{RESET}")
    print(
        f"{YELLOW}Do not forget to correct the template draft "
        + f"to your needs!{RESET}"
    )


def create_parser():
    parser = ArgumentParser(
        description="Script for the creation of template`s template"
    )

    parser.add_argument(
        "-on",
        "--output-name",
        metavar="filename",
        type=str,
        default="template",
        help="Name for the template output file.",
    )
    parser.add_argument(
        "-if",
        "--input-filename",
        metavar="filename",
        type=str,
        default=None,
        help="Relative or absolute path to the 'exchange.out' file, "
        + "including the name and extension of the file itself.",
    )
    parser.add_argument(
        "-R",
        "--R-vector",
        metavar="R1a R1b R1c",
        type=int,
        nargs="*",
        default=None,
        help="R vectors for filtering the model.",
    )
    parser.add_argument(
        "-maxd",
        "--max-distance",
        metavar="distance",
        type=float,
        default=None,
        help="(<=) Maximum distance.",
    )
    parser.add_argument(
        "-mind",
        "--min-distance",
        metavar="distance",
        type=float,
        default=None,
        help="(>=) Minimum distance.",
    )
    parser.add_argument(
        "-d",
        "--distance",
        metavar="distance",
        type=float,
        default=None,
        help="(=) Exact distance.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        default=False,
        help="Verbose output, propagates to the called methods.",
    )

    return parser



from math import cos, sin, tan

import numpy as np


class HighSymmetryPoints:
    r"""
    Generator of the high symmetry k-points.

    Parameters
    ----------
    kpoints : dict
        Dictionary of the high symmetry k-points and coordinate
        (fractions of the reciprocal vectors).

        .. code-block:: python

            kpoints = {"name" : [xb1, xb2, xb3]}
    """

    def __init__(self, kpoints=None, path=None) -> None:
        self._kpoints = kpoints
        self._path = path

        self._PLOT_LITERALS = {
            "Gamma": "$\\Gamma$",
            "M": "$M$",
            "R": "$R$",
            "X": "$X$",
            "K": "$K$",
            "L": "$L$",
            "U": "$U$",
            "W": "$W$",
            "H": "$H$",
            "P": "$P$",
            "N": "$N$",
            "A": "$A$",
            "Z": "$Z$",
            "Z1": "$Z_1$",
            "Sigma": "$\\Sigma$",
            "Sigma1": "$\\Sigma_1$",
            "Y": "$Y$",
            "Y1": "$Y_1$",
            "S": "$S$",
            "T": "$T$",
            "A1": "$A_1$",
            "X1": "$X_1$",
            "C": "$C$",
            "C1": "$C_1$",
            "D": "$D$",
            "D1": "$D_1$",
            "H1": "$H_1$",
            "L1": "$L_1$",
            "L2": "$L_2$",
            "B": "$B$",
            "B1": "$B_1$",
            "F": "$F$",
            "P1": "$P_1$",
            "P2": "$P_2$",
            "Q": "$Q$",
            "Q1": "$Q_1$",
            "E": "$E$",
            "H2": "$H_2$",
            "M1": "$M_1$",
            "M2": "$M_2$",
            "N1": "$N_1$",
            "F1": "$F_1$",
            "F2": "$F_2$",
            "F3": "$F_3$",
            "I": "$I$",
            "I1": "$I_1$",
            "X2": "$X_2$",
            "Y2": "$Y_2$",
            "Y3": "$Y_3$",
        }

    @property
    def kpoints(self):
        r"""
        Dictionary of K-points.

        .. code-block:: python

            kpoints = {Point1 : [k1, k2, k3], ...}

        where ``k1``, ``k2``, ``k3`` are
        the relative coordinates of the ``Point1``.
        """

        if self._kpoints is None:
            return {}
        else:
            return self._kpoints

    def add_kpoint(self, name, coordinates, plot_name=None):
        r"""
        Add one kpoint to the set of High symmetry kpoints.

        Parameters
        ----------
        name : str
            Name of the kpoints, which is used by the code as a key.
        coordinates : 1x3 array
            Relative coordinates of the kpoint.
        plot_name : str
            name of the kpoints to be display on the plot.
            Equal to the ``name`` by default.
        """

        coordinates = np.array(coordinates)
        if coordinates.shape != (3,):
            raise ValueError
        if self._kpoints is None:
            self._kpoints = {}
        self._kpoints[name] = coordinates

        if plot_name is None and name not in self._PLOT_LITERALS:
            self._PLOT_LITERALS[name] = name
        elif plot_name is not None:
            self._PLOT_LITERALS[name] = plot_name

    @property
    def path(self):
        r"""
        Path of K-points.

        Always is a list of lists.

        .. code-block:: python

            path = [subpath1, subpath2, ...]
            subpath1 = [hs_kpoint1, hs_kpoint2, ...]
        """

        if self._path is None:
            return []
        else:
            return self._path

    @path.setter
    def path(self, new_path):
        if new_path is None:
            self._path = []

        if isinstance(new_path, str):
            if "|-" in new_path or "-|" in new_path:
                raise ValueError("Check the format of the new path")
            subpaths = new_path.split("|")
            if subpaths[-1] == "":
                subpaths = subpaths[0:-1]
            if subpaths[0] == "":
                subpaths = subpaths[1:]

            for i in range(0, len(subpaths)):
                subpaths[i] = subpaths[i].split("-")
                for j in range(0, len(subpaths[i])):
                    if subpaths[i][j] not in self.kpoints:
                        raise ValueError(
                            f"Provided  high symmetry "
                            + f"k-point {subpaths[i][j]} is not known."
                        )
            self._path = subpaths

        if isinstance(new_path, list):
            if len(new_path) == 0:
                self._path = []
            elif isinstance(new_path[0], list):
                for i in range(0, len(new_path)):
                    for j in range(0, len(new_path[i])):
                        if new_path[i][j] not in self.kpoints:
                            raise ValueError(
                                f"Provided  high symmetry "
                                + f"k-point {new_path[i][j]} is not known."
                            )
                if len(new_path) == 1 and len(new_path[0]) == 0:
                    self._path = []
                else:
                    self._path = new_path
            elif isinstance(new_path[0], str):
                for i in range(0, len(new_path)):
                    for j in range(0, len(new_path[i])):
                        if new_path[i][j] not in self.kpoints:
                            raise ValueError(
                                f"Provided  high symmetry "
                                + f"k-point {new_path[i][j]} is not known."
                            )
                self._path = [new_path]

            else:
                raise ValueError(f"Path format is not supported: {new_path}")

    def cub(self):
        r"""Cubic (CUB, cP)"""

        kpoints = {
            "Gamma": np.array([0, 0, 0]),
            "M": np.array([1 / 2, 1 / 2, 0]),
            "R": np.array([1 / 2, 1 / 2, 1 / 2]),
            "X": np.array([0, 1 / 2, 0]),
        }

        path = [["Gamma", "X", "M", "Gamma", "R", "X"], ["M", "R"]]

        self._kpoints = kpoints
        self.path = path

        return HighSymmetryPoints(kpoints, path)

    def fcc(self):
        r"""Face-centred cubic (FCC, cF)"""

        kpoints = {
            "Gamma": np.array([0, 0, 0]),
            "K": np.array([3 / 8, 3 / 8, 3 / 4]),
            "L": np.array([1 / 2, 1 / 2, 1 / 2]),
            "U": np.array([5 / 8, 1 / 4, 5 / 8]),
            "W": np.array([1 / 2, 1 / 4, 3 / 4]),
            "X": np.array([1 / 2, 0, 1 / 2]),
        }

        path = [["Gamma", "X", "W", "K", "Gamma", "L", "U", "W", "L", "K"], ["U", "X"]]

        self._kpoints = kpoints
        self.path = path

        return HighSymmetryPoints(kpoints, path)

    def bcc(self):
        r"""Body-centered cubic (BCC, cl)"""

        kpoints = {
            "Gamma": np.array([0, 0, 0]),
            "H": np.array([1 / 2, -1 / 2, 1 / 2]),
            "P": np.array([1 / 4, 1 / 4, 1 / 4]),
            "N": np.array([0, 0, 1 / 2]),
        }

        path = [["Gamma", "H", "N", "Gamma", "P", "H"], ["P", "N"]]

        self._kpoints = kpoints
        self.path = path

        return HighSymmetryPoints(kpoints, path)

    def tet(self):
        r"""Tetragonal (TET, tP)"""

        kpoints = {
            "Gamma": np.array([0, 0, 0]),
            "A": np.array([1 / 2, 1 / 2, 1 / 2]),
            "M": np.array([1 / 2, 1 / 2, 0]),
            "R": np.array([0, 1 / 2, 1 / 2]),
            "X": np.array([0, 1 / 2, 0]),
            "Z": np.array([0, 0, 1 / 2]),
        }

        path = [
            ["Gamma", "X", "M", "Gamma", "Z", "R", "A", "Z"],
            ["X", "R"],
            ["M", "A"],
        ]

        self._kpoints = kpoints
        self.path = path

        return HighSymmetryPoints(kpoints, path)

    def bct1(self, a, c):
        r"""Body-centred tetragonal (BCT, tI),

        .. math::
            c < a
        """

        eta = (1 + c**2 / a**2) / 4

        kpoints = {
            "Gamma": np.array([0, 0, 0]),
            "M": np.array([-1 / 2, 1 / 2, 1 / 2]),
            "N": np.array([0, 1 / 2, 0]),
            "P": np.array([1 / 4, 1 / 4, 1 / 4]),
            "X": np.array([0, 0, 1 / 2]),
            "Z": np.array([eta, eta, -eta]),
            "Z1": np.array([-eta, 1 - eta, eta]),
        }

        path = [["Gamma", "X", "M", "Gamma", "Z", "P", "N", "Z1", "M"], ["X", "P"]]

        self._kpoints = kpoints
        self.path = path

        return HighSymmetryPoints(kpoints, path)

    def bct2(self, a, c):
        r"""Body-centred tetragonal (BCT, tI),

        .. math::
            c > a
        """

        eta = (1 + c**2 / a**2) / 4
        zeta = a**2 / (2 * c**2)
        kpoints = {
            "Gamma": np.array([0, 0, 0]),
            "N": np.array([0, 1 / 2, 0]),
            "P": np.array([1 / 4, 1 / 4, 1 / 4]),
            "Sigma": np.array([-eta, eta, eta]),
            "Sigma1": np.array([eta, 1 - eta, -eta]),
            "X": np.array([0, 0, 1 / 2]),
            "Y": np.array([-zeta, zeta, 1 / 2]),
            "Y1": np.array([1 / 2, 1 / 2, -zeta]),
            "Z": np.array([1 / 2, 1 / 2, -1 / 2]),
        }

        path = [
            ["gamma", "X", "Y", "Sigma", "Gamma", "Z", "Sigma1", "N", "P", "Y1", "Z"],
            ["X", "P"],
        ]

        self._kpoints = kpoints
        self.path = path

        return HighSymmetryPoints(kpoints, path)

    def orc(self):
        r"""Orthorhombic (ORC, oP)"""

        kpoints = {
            "Gamma": np.array([0, 0, 0]),
            "R": np.array([1 / 2, 1 / 2, 1 / 2]),
            "S": np.array([1 / 2, 1 / 2, 0]),
            "T": np.array([0, 1 / 2, 1 / 2]),
            "U": np.array([1 / 2, 0, 1 / 2]),
            "X": np.array([1 / 2, 0, 0]),
            "Y": np.array([0, 1 / 2, 0]),
            "Z": np.array([0, 0, 1 / 2]),
        }

        path = [
            ["Gamma", "X", "S", "Y", "Gamma", "Z", "U", "R", "T", "Z"],
            ["Y", "T"],
            ["U", "X"],
            ["S", "R"],
        ]

        self._kpoints = kpoints
        self.path = path

        return HighSymmetryPoints(kpoints, path)

    def orcf1(self, a, b, c):
        r"""Face-centred orthorhombic (ORCF, oF),

        .. math::
            \dfrac{1}{a^2} > \dfrac{1}{b^2} + \dfrac{1}{c^2}
        """

        eta = (1 + a**2 / b**2 + a**2 / c**2) / 4
        zeta = (1 + a**2 / b**2 - a**2 / c**2) / 4

        kpoints = {
            "Gamma": np.array([0, 0, 0]),
            "A": np.array([1 / 2, 1 / 2 + zeta, zeta]),
            "A1": np.array([1 / 2, 1 / 2 - zeta, 1 - zeta]),
            "L": np.array([1 / 2, 1 / 2, 1 / 2]),
            "T": np.array([1, 1 / 2, 1 / 2]),
            "X": np.array([0, eta, eta]),
            "X1": np.array([1, 1 - eta, 1 - eta]),
            "Y": np.array([1 / 2, 0, 1 / 2]),
            "Z": np.array([1 / 2, 1 / 2, 0]),
        }

        path = [
            ["Gamma", "Y", "T", "Z", "Gamma", "X", "A1", "Y"],
            ["T", "X1"],
            ["X", "A", "Z"],
            ["L", "Gamma"],
        ]

        self._kpoints = kpoints
        self.path = path

        return HighSymmetryPoints(kpoints, path)

    def orcf2(self, a, b, c):
        r"""Face-centred orthorhombic (ORCF, oF),

        .. math::
            \dfrac{1}{a^2} < \dfrac{1}{b^2} + \dfrac{1}{c^2}
        """

        eta = (1 + a**2 / b**2 + a**2 / c**2) / 4
        delta = (1 + b**2 / a**2 - b**2 / c**2) / 4
        phi = (1 + c**2 / b**2 - c**2 / a**2) / 4

        kpoints = {
            "Gamma": np.array([0, 0, 0]),
            "C": np.array([1 / 2, 1 / 2 - eta, 1 - eta]),
            "C1": np.array([1 / 2, 1 / 2 + eta, 1 - eta]),
            "D": np.array([1 / 2 - delta, 1 / 2, 1 - delta]),
            "D1": np.array([1 / 2 + delta, 1 / 2, delta]),
            "L": np.array([1 / 2, 1 / 2, 1 / 2]),
            "H": np.array([1 - phi, 1 / 2 - phi, 1 / 2]),
            "H1": np.array([phi, 1 / 2 + phi, 1 / 2]),
            "X": np.array([0, 1 / 2, 1 / 2]),
            "Y": np.array([1 / 2, 0, 1 / 2]),
            "Z": np.array([1 / 2, 1 / 2, 0]),
        }

        path = [
            ["Gamma", "Y", "C", "D", "X", "Gamma", "Z", "D1", "H", "C"],
            ["C1", "Z"],
            ["X", "H1"],
            ["H", "Y"],
            ["L", "Gamma"],
        ]

        self._kpoints = kpoints
        self.path = path

        return HighSymmetryPoints(kpoints, path)

    def orcf3(self, a, b, c):
        r"""Face-centred orthorhombic (ORCF, oF),

        .. math::
            \dfrac{1}{a^2} = \dfrac{1}{b^2} + \dfrac{1}{c^2}
        """

        eta = (1 + a**2 / b**2 + a**2 / c**2) / 4
        zeta = (1 + a**2 / b**2 - a**2 / c**2) / 4

        kpoints = {
            "Gamma": np.array([0, 0, 0]),
            "A": np.array([1 / 2, 1 / 2 + zeta, zeta]),
            "A1": np.array([1 / 2, 1 / 2 - zeta, 1 - zeta]),
            "L": np.array([1 / 2, 1 / 2, 1 / 2]),
            "T": np.array([1, 1 / 2, 1 / 2]),
            "X": np.array([0, eta, eta]),
            "X1": np.array([1, 1 - eta, 1 - eta]),
            "Y": np.array([1 / 2, 0, 1 / 2]),
            "Z": np.array([1 / 2, 1 / 2, 0]),
        }

        path = [
            ["Gamma", "Y", "T", "Z", "Gamma", "X", "A1", "Y"],
            ["X", "A", "Z"],
            ["L", "Gamma"],
        ]

        self._kpoints = kpoints
        self.path = path

        return HighSymmetryPoints(kpoints, path)

    def orci(self, a, b, c):
        r"""Body-centred orthorhombic (ORCI, oI)"""

        eta = (1 + a**2 / c**2) / 4
        zeta = (1 + b**2 / c**2) / 4
        delta = (b**2 - a**2) / (4 * c**2)
        mu = (a**2 + b**2) / (4 * c**2)

        kpoints = {
            "Gamma": np.array([0, 0, 0]),
            "L": np.array([-mu, mu, 1 / 2 - delta]),
            "L1": np.array([mu, -mu, 1 / 2 + delta]),
            "L2": np.array([1 / 2 - delta, 1 / 2 + delta, -mu]),
            "R": np.array([0, 1 / 2, 0]),
            "S": np.array([1 / 2, 0, 0]),
            "T": np.array([0, 0, 1 / 2]),
            "W": np.array([1 / 4, 1 / 4, 1 / 4]),
            "X": np.array([-zeta, zeta, zeta]),
            "X1": np.array([zeta, 1 - zeta, -zeta]),
            "Y": np.array([eta, -eta, eta]),
            "Y1": np.array([1 - eta, eta, -eta]),
            "Z": np.array([1 / 2, 1 / 2, -1 / 2]),
        }

        path = [
            ["Gamma", "X", "L", "T", "W", "R", "X1", "Z", "Gamma", "Y", "S", "W"],
            ["L1", "Y"],
            ["Y1", "Z"],
        ]

        self._kpoints = kpoints
        self.path = path

        return HighSymmetryPoints(kpoints, path)

    def orcc(self, a, b):
        r"""C-centred orthorhombic (ORCC, oS)"""

        zeta = (1 + a**2 / b**2) / 4

        kpoints = {
            "Gamma": np.array([0, 0, 0]),
            "A": np.array([zeta, zeta, 1 / 2]),
            "A1": np.array([-zeta, 1 - zeta, 1 / 2]),
            "R": np.array([0, 1 / 2, 1 / 2]),
            "S": np.array([0, 1 / 2, 0]),
            "T": np.array([-1 / 2, 1 / 2, 1 / 2]),
            "X": np.array([zeta, zeta, 0]),
            "X1": np.array([-zeta, 1 - zeta, 0]),
            "Y": np.array([-1 / 2, 1 / 2, 0]),
            "Z": np.array([0, 0, 1 / 2]),
        }

        path = [
            ["Gamma", "X", "S", "R", "A", "Z", "Gamma", "Y", "X1", "A1", "T", "Y"],
            ["Z", "T"],
        ]

        self._kpoints = kpoints
        self.path = path

        return HighSymmetryPoints(kpoints, path)

    def hex(self):
        r"""Hexagonal (HEX, hP)"""

        kpoints = {
            "Gamma": np.array([0, 0, 0]),
            "A": np.array([0, 0, 1 / 2]),
            "H": np.array([1 / 3, 1 / 3, 1 / 2]),
            "K": np.array([1 / 3, 1 / 3, 0]),
            "L": np.array([1 / 2, 0, 1 / 2]),
            "M": np.array([1 / 2, 0, 0]),
        }

        path = [
            ["Gamma", "M", "K", "Gamma", "A", "L", "H", "A"],
            ["L", "M"],
            ["K", "H"],
        ]

        self._kpoints = kpoints
        self.path = path

        return HighSymmetryPoints(kpoints, path)

    def rhl1(self, alpha):
        r"""Rhombohedral (RHL, hR),

        .. math::
            \alpha < 90^{\circ}
        """

        eta = (1 + 4 * cos(alpha)) / (2 + 4 * cos(alpha))
        nu = 3 / 4 - eta / 2

        kpoints = {
            "Gamma": np.array([0, 0, 0]),
            "B": np.array([eta, 1 / 2, 1 - eta]),
            "B1": np.array([1 / 2, 1 - eta, eta - 1]),
            "F": np.array([1 / 2, 1 / 2, 0]),
            "L": np.array([1 / 2, 0, 0]),
            "L1": np.array([0, 0, -1 / 2]),
            "P": np.array([eta, nu, nu]),
            "P1": np.array([1 - nu, 1 - nu, 1 - eta]),
            "P2": np.array([nu, nu, eta - 1]),
            "Q": np.array([1 - nu, nu, 0]),
            "X": np.array([nu, 0, -nu]),
            "Z": np.array([1 / 2, 1 / 2, 1 / 2]),
        }

        path = [
            ["Gamma", "L", "B1"],
            ["B", "Z", "Gamma", "X"],
            ["Q", "F", "P1", "Z"],
            ["L", "P"],
        ]

        self._kpoints = kpoints
        self.path = path

        return HighSymmetryPoints(kpoints, path)

    def rhl2(self, alpha):
        r"""Rhombohedral (RHL, hR),

        .. math::
            \alpha > 90^{\circ}
        """

        eta = 1 / (2 * tan(alpha / 2) ** 2)
        nu = 3 / 4 - eta / 2

        kpoints = {
            "Gamma": np.array([0, 0, 0]),
            "F": np.array([1 / 2, -1 / 2, 0]),
            "L": np.array([1 / 2, 0, 0]),
            "P": np.array([1 - nu, -nu, 1 - nu]),
            "P1": np.array([nu, nu - 1, nu - 1]),
            "Q": np.array([eta, eta, eta]),
            "Q1": np.array([1 - eta, -eta, -eta]),
            "Z": np.array([1 / 2, -1 / 2, 1 / 2]),
        }

        path = [["Gamma", "P", "Z", "Q", "Gamma", "F", "P1", "Q1", "L", "Z"]]

        self._kpoints = kpoints
        self.path = path

        return HighSymmetryPoints(kpoints, path)

    def mcl(self, b, c, alpha):
        r"""Monoclinic (MCL, mP)"""

        eta = (1 - b * cos(alpha) / c) / (2 * sin(alpha) ** 2)
        nu = 1 / 2 - eta * c * cos(alpha) / b

        kpoints = {
            "Gamma": np.array([0, 0, 0]),
            "A": np.array([1 / 2, 1 / 2, 0]),
            "C": np.array([0, 1 / 2, 1 / 2]),
            "D": np.array([1 / 2, 0, 1 / 2]),
            "D1": np.array([1 / 2, 0, -1 / 2]),
            "E": np.array([1 / 2, 1 / 2, 1 / 2]),
            "H": np.array([0, eta, 1 - nu]),
            "H1": np.array([0, 1 - eta, nu]),
            "H2": np.array([0, eta, -nu]),
            "M": np.array([1 / 2, eta, 1 - nu]),
            "M1": np.array([1 / 2, 1 - eta, nu]),
            "M2": np.array([1 / 2, eta, -nu]),
            "X": np.array([0, 1 / 2, 0]),
            "Y": np.array([0, 0, 1 / 2]),
            "Y1": np.array([0, 0, -1 / 2]),
            "Z": np.array([1 / 2, 0, 0]),
        }

        path = [
            ["Gamma", "Y", "H", "C", "E", "M1", "A", "X", "H1"],
            ["M", "D", "Z"],
            ["Y", "D"],
        ]

        self._kpoints = kpoints
        self.path = path

        return HighSymmetryPoints(kpoints, path)

    def mclc1(self, a, b, c, alpha):
        r"""C-centred monoclinic (MCLC, mS),

        .. math::
            k_{\gamma} > 90^{\circ}
        """

        zeta = (2 - b * cos(alpha) / c) / (4 * sin(alpha) ** 2)
        eta = 1 / 2 + 2 * eta * c * cos(alpha) / b
        psi = 3 / 4 - a**2 / (4 * b**2 * sin(alpha) ** 2)
        phi = psi + (3 / 4 - psi) * b * cos(alpha) / c

        kpoints = {
            "Gamma": np.array([0, 0, 0]),
            "N": np.array([1 / 2, 0, 0]),
            "N1": np.array([0, -1 / 2, 0]),
            "F": np.array([1 - zeta, 1 - zeta, 1 - eta]),
            "F1": np.array([zeta, zeta, eta]),
            "F2": np.array([-zeta, -zeta, 1 - eta]),
            "F3": np.array([1 - zeta, -zeta, 1 - eta]),
            "I": np.array([phi, 1 - phi, 1 / 2]),
            "I1": np.array([1 - phi, phi - 1, 1 / 2]),
            "L": np.array([1 / 2, 1 / 2, 1 / 2]),
            "M": np.array([1 / 2, 0, 1 / 2]),
            "X": np.array([1 - psi, psi - 1, 0]),
            "X1": np.array([psi, 1 - psi, 0]),
            "X2": np.array([psi - 1, -psi, 0]),
            "Y": np.array([1 / 2, 1 / 2, 0]),
            "Y1": np.array([-1 / 2, -1 / 2, 0]),
            "Z": np.array([0, 0, 1 / 2]),
        }

        path = [
            ["Gamma", "Y", "F", "L", "I"],
            ["I1", "Z", "F1"],
            ["Y", "X1"],
            ["X", "Gamma", "N"],
            ["M", "Gamma"],
        ]

        self._kpoints = kpoints
        self.path = path

        return HighSymmetryPoints(kpoints, path)

    def mclc2(self, a, b, c, alpha):
        r"""C-centred monoclinic (MCLC, mS),

        .. math::
            k_{\gamma} = 90^{\circ}
        """

        zeta = (2 - b * cos(alpha) / c) / (4 * sin(alpha) ** 2)
        eta = 1 / 2 + 2 * eta * c * cos(alpha) / b
        psi = 3 / 4 - a**2 / (4 * b**2 * sin(alpha) ** 2)
        phi = psi + (3 / 4 - psi) * b * cos(alpha) / c

        kpoints = {
            "Gamma": np.array([0, 0, 0]),
            "N": np.array([1 / 2, 0, 0]),
            "N1": np.array([0, -1 / 2, 0]),
            "F": np.array([1 - zeta, 1 - zeta, 1 - eta]),
            "F1": np.array([zeta, zeta, eta]),
            "F2": np.array([-zeta, -zeta, 1 - eta]),
            "F3": np.array([1 - zeta, -zeta, 1 - eta]),
            "I": np.array([phi, 1 - phi, 1 / 2]),
            "I1": np.array([1 - phi, phi - 1, 1 / 2]),
            "L": np.array([1 / 2, 1 / 2, 1 / 2]),
            "M": np.array([1 / 2, 0, 1 / 2]),
            "X": np.array([1 - psi, psi - 1, 0]),
            "X1": np.array([psi, 1 - psi, 0]),
            "X2": np.array([psi - 1, -psi, 0]),
            "Y": np.array([1 / 2, 1 / 2, 0]),
            "Y1": np.array([-1 / 2, -1 / 2, 0]),
            "Z": np.array([0, 0, 1 / 2]),
        }

        path = [["Gamma", "Y", "F", "L", "I"], ["I1", "Z", "F1"], ["N", "Gamma", "M"]]

        self._kpoints = kpoints
        self.path = path

        return HighSymmetryPoints(kpoints, path)

    def mclc3(self, a, b, c, alpha):
        r"""C-centred monoclinic (MCLC, mS),

        .. math::
            k_{\gamma} < 90^{\circ},
            \dfrac{b\cos(\alpha)}{c} + \dfrac{b^2\sin(\alpha)^2}{a^2} < 1
        """

        mu = (1 + b**2 / a**2) / 4
        delta = b * c * cos(alpha) / (2 * a**2)
        zeta = mu - 1 / 4 + (1 - b * cos(alpha) / c) / (4 * sin(alpha) ^ 2)
        eta = 1 / 2 + 2 * zeta * c * cos(alpha) / b
        phi = 1 + zeta - 2 * mu
        psi = eta - 2 * delta

        kpoints = {
            "Gamma": np.array([0, 0, 0]),
            "F": np.array([1 - phi, 1 - phi, 1 - psi]),
            "F1": np.array([phi, phi - 1, psi]),
            "F2": np.array([1 - phi, -phi, 1 - psi]),
            "H": np.array([zeta, zeta, eta]),
            "H1": np.array([1 - zeta, -zeta, 1 - eta]),
            "H2": np.array([-zeta, -zeta, 1 - eta]),
            "I": np.array([1 / 2, -1 / 2, 1 / 2]),
            "M": np.array([1 / 2, 0, 1 / 2]),
            "N": np.array([1 / 2, 0, 0]),
            "N1": np.array([0, -1 / 2, 0]),
            "X": np.array([1 / 2, -1 / 2, 0]),
            "Y": np.array([mu, mu, delta]),
            "Y1": np.array([1 - mu, -mu, -delta]),
            "Y2": np.array([-mu, -mu, -delta]),
            "Y3": np.array([mu, mu - 1, delta]),
            "Z": np.array([0, 0, 1 / 2]),
        }

        path = [
            ["Gamma", "Y", "F", "H", "Z", "I", "F1"],
            ["H1", "Y1", "X", "Gamma", "N"],
            ["M", "Gamma"],
        ]

        self._kpoints = kpoints
        self.path = path

        return HighSymmetryPoints(kpoints, path)

    def mclc4(self, a, b, c, alpha):
        r"""C-centred monoclinic (MCLC, mS),

        .. math::
            k_{\gamma} < 90^{\circ},
            \dfrac{b\cos(\alpha)}{c} + \dfrac{b^2\sin(\alpha)^2}{a^2} = 1
        """

        mu = (1 + b**2 / a**2) / 4
        delta = b * c * cos(alpha) / (2 * a**2)
        zeta = mu - 1 / 4 + (1 - b * cos(alpha) / c) / (4 * sin(alpha) ^ 2)
        eta = 1 / 2 + 2 * zeta * c * cos(alpha) / b
        phi = 1 + zeta - 2 * mu
        psi = eta - 2 * delta

        kpoints = {
            "Gamma": np.array([0, 0, 0]),
            "F": np.array([1 - phi, 1 - phi, 1 - psi]),
            "F1": np.array([phi, phi - 1, psi]),
            "F2": np.array([1 - phi, -phi, 1 - psi]),
            "H": np.array([zeta, zeta, eta]),
            "H1": np.array([1 - zeta, -zeta, 1 - eta]),
            "H2": np.array([-zeta, -zeta, 1 - eta]),
            "I": np.array([1 / 2, -1 / 2, 1 / 2]),
            "M": np.array([1 / 2, 0, 1 / 2]),
            "N": np.array([1 / 2, 0, 0]),
            "N1": np.array([0, -1 / 2, 0]),
            "X": np.array([1 / 2, -1 / 2, 0]),
            "Y": np.array([mu, mu, delta]),
            "Y1": np.array([1 - mu, -mu, -delta]),
            "Y2": np.array([-mu, -mu, -delta]),
            "Y3": np.array([mu, mu - 1, delta]),
            "Z": np.array([0, 0, 1 / 2]),
        }

        path = [
            ["Gamma", "Y", "F", "H", "Z", "I"],
            ["H1", "Y1", "X", "Gamma", "N"],
            ["M", "Gamma"],
        ]

        self._kpoints = kpoints
        self.path = path

        return HighSymmetryPoints(kpoints, path)

    def mclc5(self, a, b, c, alpha):
        r"""C-centred monoclinic (MCLC, mS),

        .. math::
            k_{\gamma} < 90^{\circ},
            \dfrac{b\cos(\alpha)}{c} + \dfrac{b^2\sin(\alpha)^2}{a^2} > 1
        """

        zeta = (b**2 / a**2 + (1 - b * cos(alpha) / c) / sin(alpha) ** 2) / 4
        mu = zeta / 2 + b**2 / (4 * a**2) - b * c * cos(alpha) / (2 * a**2)
        nu = 2 * mu - zeta
        rho = 1 - zeta * a**2 / b**2
        omega = (
            (4 * nu - 1 - b**2 * sin(alpha) ** 2 / a**2) * c / (2 * b * cos(alpha))
        )
        eta = 1 / 2 + 2 * zeta * c * cos(alpha) / b
        delta = zeta * c * cos(alpha) / b + omega / 2 - 1 / 4

        kpoints = {
            "Gamma": np.array([0, 0, 0]),
            "F": np.array([nu, nu, omega]),
            "F1": np.array([1 - nu, 1 - nu, 1 - omega]),
            "F2": np.array([nu, nu - 1, omega]),
            "H": np.array([zeta, zeta, eta]),
            "H1": np.array([1 - zeta, -zeta, 1 - eta]),
            "H2": np.array([-zeta, -zeta, 1 - eta]),
            "I": np.array([rho, 1 - rho, 1 / 2]),
            "I1": np.array([1 - rho, rho - 1, 1 / 2]),
            "L": np.array([1 / 2, 1 / 2, 1 / 2]),
            "M": np.array([1 / 2, 0, 1 / 2]),
            "N": np.array([1 / 2, 0, 0]),
            "N1": np.array([0, -1 / 2, 0]),
            "X": np.array([1 / 2, -1 / 2, 0]),
            "Y": np.array([mu, mu, delta]),
            "Y1": np.array([1 - mu, -mu, -delta]),
            "Y2": np.array([-mu, -mu, -delta]),
            "Y3": np.array([mu, mu - 1, delta]),
            "Z": np.array([0, 0, 1 / 2]),
        }

        path = [
            ["Gamma", "Y", "F", "L", "I"],
            ["I1", "Z", "H", "F1"],
            ["H1", "Y1", "X", "Gamma", "N"],
            ["M", "Gamma"],
        ]

        self._kpoints = kpoints
        self.path = path

        return HighSymmetryPoints(kpoints, path)

    def tri1a(self, a):
        r"""Triclinic (TRI, aP),

        .. math::
            k_{\alpha} > 90^{\circ},
            k_{\beta} > 90^{\circ},
            k_{\gamma} > 90^{\circ}
        .. math::
            k_{\gamma} = \min(k_{\alpha}, k_{\beta}, k_{\gamma})
        """

        kpoints = {
            "Gamma": np.array([0, 0, 0]),
            "L": np.array([1 / 2, 1 / 2, 0]),
            "M": np.array([0, 1 / 2, 1 / 2]),
            "N": np.array([1 / 2, 0, 1 / 2]),
            "R": np.array([1 / 2, 1 / 2, 1 / 2]),
            "X": np.array([1 / 2, 0, 0]),
            "Y": np.array([0, 1 / 2, 0]),
            "Z": np.array([0, 0, 1 / 2]),
        }

        path = [
            ["X", "Gamma", "Y"],
            ["L", "Gamma", "Z"],
            ["N", "Gamma", "M"],
            "R",
            "Gamma",
        ]

        self._kpoints = kpoints
        self.path = path

        return HighSymmetryPoints(kpoints, path)

    def tri1b(self, a):
        r"""Triclinic (TRI, aP),

        .. math::
            k_{\alpha} < 90^{\circ},
            k_{\beta} < 90^{\circ},
            k_{\gamma} < 90^{\circ}
        .. math::
            k_{\gamma} = \max(k_{\alpha}, k_{\beta}, k_{\gamma})
        """

        kpoints = {
            "Gamma": np.array([0, 0, 0]),
            "L": np.array([1 / 2, -1 / 2, 0]),
            "M": np.array([0, 0, 1 / 2]),
            "N": np.array([-1 / 2, -1 / 2, 1 / 2]),
            "R": np.array([0, -1 / 2, 1 / 2]),
            "X": np.array([0, -1 / 2, 0]),
            "Y": np.array([1 / 2, 0, 0]),
            "Z": np.array([-1 / 2, 0, 1 / 2]),
        }

        path = [
            ["X", "Gamma", "Y"],
            ["L", "Gamma", "Z"],
            ["N", "Gamma", "M"],
            "R",
            "Gamma",
        ]

        self._kpoints = kpoints
        self.path = path

        return HighSymmetryPoints(kpoints, path)

    def tri2a(self, a):
        r"""Triclinic (TRI, aP),

        .. math::
            k_{\alpha} > 90^{\circ},
            k_{\beta} > 90^{\circ},
            k_{\gamma} = 90^{\circ}
        """

        kpoints = {
            "Gamma": np.array([0, 0, 0]),
            "L": np.array([1 / 2, 1 / 2, 0]),
            "M": np.array([0, 1 / 2, 1 / 2]),
            "N": np.array([1 / 2, 0, 1 / 2]),
            "R": np.array([1 / 2, 1 / 2, 1 / 2]),
            "X": np.array([1 / 2, 0, 0]),
            "Y": np.array([0, 1 / 2, 0]),
            "Z": np.array([0, 0, 1 / 2]),
        }

        path = [
            ["X", "Gamma", "Y"],
            ["L", "Gamma", "Z"],
            ["N", "Gamma", "M"],
            "R",
            "Gamma",
        ]

        self._kpoints = kpoints
        self.path = path

        return HighSymmetryPoints(kpoints, path)

    def tri2b(self, a):
        r"""Triclinic (TRI, aP),

        .. math::
            k_{\alpha} < 90^{\circ},
            k_{\beta} < 90^{\circ},
            k_{\gamma} = 90^{\circ}
        """

        kpoints = {
            "Gamma": np.array([0, 0, 0]),
            "L": np.array([1 / 2, -1 / 2, 0]),
            "M": np.array([0, 0, 1 / 2]),
            "N": np.array([-1 / 2, -1 / 2, 1 / 2]),
            "R": np.array([0, -1 / 2, 1 / 2]),
            "X": np.array([0, -1 / 2, 0]),
            "Y": np.array([1 / 2, 0, 0]),
            "Z": np.array([-1 / 2, 0, 1 / 2]),
        }

        path = [
            ["X", "Gamma", "Y"],
            ["L", "Gamma", "Z"],
            ["N", "Gamma", "M"],
            "R",
            "Gamma",
        ]

        self._kpoints = kpoints
        self.path = path

        return HighSymmetryPoints(kpoints, path)



r"""
Collection of small routines and constants, 
which may be used across the whole package.
"""

from math import asin, pi, sqrt

import numpy as np

# Terminal colours
BLACK = "\u001b[30m"
"""
ANSI escape code for black color of the text.
"""
RED = "\u001b[31m"
"""
ANSI escape code for red color of the text.
"""
GREEN = "\u001b[32m"
"""
ANSI escape code for green color of the text.
"""
YELLOW = "\u001b[33m"
"""
ANSI escape code for yellow color of the text.
"""
BLUE = "\u001b[34m"
"""
ANSI escape code for blue color of the text.
"""
MAGENTA = "\u001b[35m"
"""
ANSI escape code for magenta color of the text.
"""
CYAN = "\u001b[36m"
"""
ANSI escape code for cyan color of the text.
"""
WHITE = "\u001b[37m"
"""
ANSI escape code for white color of the text.
"""
RESET = "\u001b[0m"
"""
ANSI escape code for resetting color to default.
"""
WARNING = YELLOW
"""
ANSI escape code for warnings.
"""
OK = GREEN
"""
ANSI escape code for ok messages.
"""
ERROR = RED
"""
ANSI escape code for errors.
"""


def get_named_colours(colour: str):
    colour_dict = {
        "black": BLACK,
        "red": RED,
        "green": GREEN,
        "yellow": YELLOW,
        "blue": BLUE,
        "magenta": MAGENTA,
        "cyan": CYAN,
        "white": WHITE,
    }
    colour = colour.lower()
    if colour not in colour_dict:
        return ""
    return colour_dict[colour]


def get_256_colours(n):
    r"""
    ANSI escape codes for terminal color with 256-colours support
    (see: :ANSI:`wiki <>`).

    Parameters
    ----------
    n : int
        Integer from 0 to 255 to be mapped to colours.

    Returns
    -------
    str
        string with the ANSI escape code.
    """

    if type(n) != int or not 0 <= n <= 255:
        raise ValueError(
            f"Integer n have to be in range 0 <= n <= 255. "
            f"You provided n = {n}, type<n> = {type(n)}"
        )
    return f"\033[38:5:{n}m"


def cprint(*args, colour=None, **kwargs):
    if isinstance(colour, int):
        colour = get_256_colours(colour % 256)
    elif isinstance(colour, str):
        colour = get_named_colours(colour)
    else:
        raise ValueError("Colour must be a string or integer")
    if colour is not None:
        print(colour, end="")
    print(*args, **kwargs)
    if colour is not None:
        print(RESET, end="")


def atom_mark_to_latex(mark):
    r"""
    Latexifier for atom marks.

    Cr12 -> Cr\ :sub:`12`\.

    Parameters
    ----------
    mark : str
        Mark of atom.

    Returns
    -------
    new_mark : str
        Latex version of the mark.
    """
    numbers = "0123456789"
    new_mark = "$"
    insert_underline = False
    for symbol in mark:
        if symbol in numbers and not insert_underline:
            insert_underline = True
            new_mark += "_{"
        new_mark += symbol
    new_mark += "}$"
    return new_mark


def rot_angle(x, y, dummy=False):
    r"""
    Rotational angle from 2D vector.

    Mathematically positive => counterclockwise.
    From [0 to 360)

    Parameters
    ----------
    x : float or int
        x coordinate of a vector.
    y : float or int
        y coordinate of a vector.
    """
    # rot_cos = x / (x ** 2 + y ** 2) ** 0.5
    # rot_angle = m.acos(rot_cos) / m.pi * 180
    try:
        sin = abs(y) / sqrt(x**2 + y**2)
    except ZeroDivisionError:
        raise ValueError("Angle is ill defined (x = y = 0).")
    if x > 0:
        if y > 0:
            return asin(sin) / pi * 180
        elif y == 0:
            return 0
        elif y < 0:
            if not dummy:
                return -asin(sin) / pi * 180
            return 360 - asin(sin) / pi * 180
    elif x == 0:
        if y > 0:
            return 90
        elif y == 0:
            raise ValueError("Angle is ill defined (x = y = 0).")
        elif y < 0:
            if not dummy:
                return 90
            return 270
    elif x < 0:
        if y > 0:
            if not dummy:
                return -asin(sin) / pi * 180
            return 180 - asin(sin) / pi * 180
        elif y == 0:
            if not dummy:
                return 0
            return 180
        elif y < 0:
            if not dummy:
                return asin(sin) / pi * 180
            return 180 + asin(sin) / pi * 180


def absolute_to_relative(cell, x, y, z):
    r"""
    Compute relative coordinates with respect to the unit cell.

    Parameters
    ----------
    cell : 3 x 3 array
        Lattice vectors.
    x : float
        x coordinate.
    y : float
        y coordinate.
    z : float
        z coordinate.

    Returns
    -------
    relative : 1 x 3 array
        Relative coordinate.
    """

    a = np.array(cell[0], dtype=float)
    b = np.array(cell[1], dtype=float)
    c = np.array(cell[2], dtype=float)
    v = np.array([x, y, z], dtype=float)
    if (v == np.zeros(3)).all():
        return np.zeros(3)
    B = np.array([np.dot(a, v), np.dot(b, v), np.dot(c, v)])
    A = np.array(
        [
            [np.dot(a, a), np.dot(a, b), np.dot(a, c)],
            [np.dot(b, a), np.dot(b, b), np.dot(b, c)],
            [np.dot(c, a), np.dot(c, b), np.dot(c, c)],
        ]
    )
    relative = np.linalg.solve(A, B)
    return relative



r"""
Input-output for the files related with this package.
"""

from rad_tools.exchange.template import ExchangeTemplate


def read_template(filename):
    r"""
    Read template from the template file.

    Parameters
    ----------
    filename : str
        Path to the template file.

    Returns
    -------
    template : :py:class:`.ExchangeTemplate`
        Exchange template.
    """

    # Constants
    major_sep = "=" * 20
    minor_sep = "-" * 20
    neighbors_flag = "Neighbors template:"

    template = ExchangeTemplate()

    with open(filename) as file:
        line = True
        while line:
            line = file.readline()

            # Read the neighbors names
            if line and neighbors_flag in line:
                while line and major_sep not in line:
                    if line and minor_sep in line:
                        line = file.readline()
                        name = line.split()[0]
                        try:
                            latex_name = line.split()[1]
                        except IndexError:
                            latex_name = name
                        template.names[name] = []
                        template.latex_names[name] = latex_name
                        line = file.readline()
                        while line and minor_sep not in line and major_sep not in line:
                            atom1 = line.split()[0]
                            atom2 = line.split()[1]

                            R = tuple(map(int, line.split()[2:]))
                            template.names[name].append((atom1, atom2, R))
                            line = file.readline()
                    if line and major_sep in line:
                        break
                    if line and minor_sep not in line:
                        line = file.readline()
    return template



r"""
DOS
"""
import re
from os import walk
from os.path import abspath, join

import matplotlib.pyplot as plt
import numpy as np

from rad_tools.dos.pdos import PDOSQE


class DOSQE:
    r"""
    Analyse the folder with the Quantum Espresso results and store all information.

    Parameters
    ----------
    seedname : str
        Seedname for the Quantum Espresso output files.
    input_folder : str
        Path to the folder with Quantum Espresso output files.
    energy_window : tuple, default None
        Energy limits, necessary for correct plotting.

    Attributes
    ----------
    seedname : str
        Seedname for the Quantum Espresso output files.
    energy : array
        Energy values of the DOS/PDOS.
    k_resolved : bool
    case : int
    casename : str
        Human readable name of the case base on ``case``.
    nkpoints : int
        Number of k points for k resolved DOS. 1 if DOS is not k resolved.
    nepoints : int
        Number of energy points.
    filenames : list
    atoms : list
    """

    def __init__(self, seedname: str, input_path: str, energy_window=None) -> None:
        self.seedname = seedname
        self._input_path = input_path

        # Case and whether DOS is k resolved
        self._k_resolved = None
        self._case = None

        # Detect number of energy and k points and read energy values
        self.nkpoints = None
        self.nepoints = None
        self.energy = None
        self._extract_energy()

        self._filenames = None

        self._atoms = {}
        self._wfcs = {}
        self._decompose_filenames()

        if energy_window is None:
            self.energy_window = (0, -1)
        else:
            i, j = 0, 0
            for i_e, e in enumerate(self.energy):
                if e <= energy_window[0]:
                    i += 1
                if e <= energy_window[1]:
                    j += 1
            self.energy_window = (i, min(j + 1, i_e))
        self.energy = self.energy[self.energy_window[0] : self.energy_window[1]]

    @property
    def case(self):
        r"""
        Detects case of the DOS calculations.

        Detects the case when called for the first time.

        There are 4 cases:

        #. Collinear
        #. Collinear, spin-polarized
        #. Non-collinear, non-spin-orbit
        #. Non-collinear, spin-orbit

        Headers of "seedname".pdos_tot with according to the
        :projwfc:`QE <>`:

        * Collinear:
            E DOS(E) PDOS(E)

        * Collinear, spin-polarized:
            E DOSup(E) DOSdw(E) PDOSup(E) PDOSdw(E)

        * Non-collinear, non-spin-orbit:
            E DOS(E) PDOSup(E) PDOSdw(E)

        * Non-collinear, spin-orbit:
            E DOS(E) PDOS(E)
        """

        if self._case is None:
            # Detect the case
            with open(join(f"{self._input_path}", f"{self.seedname}.pdos_tot")) as file:
                header = file.readline().lower().split()

            if "dos(e)" in header and "pdos(e)" in header:
                self._case = 1
            if "dos(e)" in header and "pdosup(e)" in header and "pdosdw(e)" in header:
                self._case = 3
            if (
                "dosup(e)" in header
                and "dosdw(e)" in header
                and "pdosup(e)" in header
                and "pdosdw(e)" in header
            ):
                self._case = 2
            if self._case == 1:
                pattern = ".pdos_atm#[0-9]*\\([a-zA-Z]*\\)_wfc#[0-9]*\\([spdf.]*\\)"
                filename = self.filenames[0]
                if not re.fullmatch(
                    f"{re.escape(self.seedname)}{pattern}",
                    filename,
                ):
                    self._case = 4
            if self._case is None:
                raise RuntimeError(
                    "Unable to detect case, analysed header:\n" + f"{header}"
                )

        return self._case

    @property
    def casename(self):
        r"""
        Human-readable name of the case.

        Returns
        -------
        case_name : str
            Name of the case.
        """
        case_names = [
            "collinear, spin-unpolarized",
            "collinear, spin-polarized",
            "non-collinear, non spin-orbit",
            "non-collinear, spin-orbit",
        ]
        return case_names[self.case - 1]

    @property
    def k_resolved(self):
        r"""
        Whether DOS is k-resolved.
        """

        if self._k_resolved is None:
            # Check for k-resolved
            with open(join(f"{self._input_path}", f"{self.seedname}.pdos_tot")) as file:
                self._k_resolved = "ik" in file.readline().lower().split()

        return self._k_resolved

    def _extract_energy(self):
        dos = np.loadtxt(
            join(f"{self._input_path}", f"{self.seedname}.pdos_tot"), skiprows=1
        ).T
        if self.k_resolved:
            self.nkpoints = int(dos[0][-1])
            self.nepoints = len(dos[0]) // self.nkpoints
            self.energy = dos[1][0 : self.nepoints]
        else:
            self.nkpoints = 1
            self.nepoints = len(dos[0])
            self.energy = dos[0]

    @property
    def filenames(self):
        r"""
        List of filenames (without "seedname".pdos_tot).
        """
        if self._filenames is None:
            pattern = ".pdos_atm#[0-9]*\\([a-zA-Z0-9]*\\)_wfc#[0-9]*\\([spdf_0-9j.]*\\)"

            # Get list of files in the folder
            all_filenames = []
            for dirpath, dirnames, filenames in walk(self._input_path):
                all_filenames.extend(filenames)
                break

            self._filenames = []
            # Search
            for filename in all_filenames:
                if re.fullmatch(
                    f"{re.escape(self.seedname)}{pattern}",
                    filename,
                ):
                    self._filenames.append(filename)

        return self._filenames

    def _decompose_filenames(self):
        r"""
        Decompose filenames and extract information about atoms and projectors.

        Note
        ----
        atoms : dist
            Dictionary with the atom labels and their numbers.

            .. code-block:: python

                atoms = {atom1: [n_1, n_2, n_3, ...]}
        wfcs : dict
            Dictionary of projectors functions and their numbers.

            .. code-block:: python

                wfcs = {(atom1, atom1_number): (wfc_label, wfc_number), ...}
        """

        for filename in self.filenames:
            # Detect names and numbers
            atom_number = int(filename.split(".pdos_atm#")[1].split("(")[0])
            atom_symbol = filename.split("(")[1].split(")")[0]
            wfc_number = int(filename.split(")_wfc#")[1].split("(")[0])
            wfc_symbol = filename.split(")_wfc#")[1].split("(")[1].split(")")[0]

            if atom_symbol not in self._atoms:
                self._atoms[atom_symbol] = []
            self._atoms[atom_symbol].append(atom_number)

            if (atom_symbol, atom_number) not in self._wfcs:
                self._wfcs[(atom_symbol, atom_number)] = []
            self._wfcs[(atom_symbol, atom_number)].append((wfc_symbol, wfc_number))

        # Sort entries
        for atom in self._atoms:
            self._atoms[atom] = list(set(self._atoms[atom]))
            self._atoms[atom].sort()
        for key in self._wfcs:
            self._wfcs[key].sort(key=lambda x: x[1])

    def __iter__(self):
        return DOSIterator(self)

    def __contains__(self, item):
        atom, atom_number, wfc, wfc_number = item
        return (
            atom in self.atoms
            and atom_number in self.atom_numbers(atom)
            and (wfc, wfc_number) in self.wfcs(atom, atom_number)
        )

    def total_dos(self, squeeze=False):
        r"""
        Total density of states.

        Density of states, computed directly from plane-wave basis set.

        If DOS is k resolved, then sum over k points.

        Parameters
        ----------
        squeeze : bool, default False
            Whenever to sum over k points. Ignored if DOS is not k resolved.

        Returns
        -------
        total_dos : array
            Shape :math:`(2, n_e)` if DOS is collinear, spin-polarized;
            :math:`(n_e)` otherwise.
        """

        dos = np.loadtxt(
            join(self._input_path, f"{self.seedname}.pdos_tot"), skiprows=1
        ).T

        if self.case == 2:
            if self.k_resolved:
                if squeeze:
                    return (
                        np.sum(
                            dos[2:4].reshape((2, self.nkpoints, self.nepoints)), axis=1
                        )
                        / self.nkpoints
                    )[:, self.energy_window[0] : self.energy_window[1]]
                return dos[2:4].reshape((2, self.nkpoints, self.nepoints))[
                    :, :, self.energy_window[0] : self.energy_window[1]
                ]
            return dos[1:3][:, self.energy_window[0] : self.energy_window[1]]

        if self.k_resolved:
            if squeeze:
                return (
                    np.sum(dos[2].reshape((self.nkpoints, self.nepoints)), axis=0)
                    / self.nkpoints
                )[self.energy_window[0] : self.energy_window[1]]
            return dos[2].reshape((self.nkpoints, self.nepoints))[
                :, self.energy_window[0] : self.energy_window[1]
            ]
        return dos[1][self.energy_window[0] : self.energy_window[1]]

    def total_pdos(self, squeeze=False):
        r"""
        Total partial density of states.

        If DOS is k resolved, then sum over k points.

        Parameters
        ----------
        squeeze : bool, default False
            Whenever to sum over k points. Ignored if DOS is not k resolved.

        Returns
        -------
        total_pdos : array
            Shape :math:`(2, n_e)` if PDOS is collinear, spin-polarized;
            :math:`(n_e)` otherwise.
        """

        dos = np.loadtxt(
            join(f"{self._input_path}", f"{self.seedname}.pdos_tot"), skiprows=1
        ).T

        if self.case == 2:
            if self.k_resolved:
                if squeeze:
                    return (
                        np.sum(
                            dos[4:6].reshape((2, self.nkpoints, self.nepoints)), axis=1
                        )
                        / self.nkpoints
                    )[:, self.energy_window[0] : self.energy_window[1]]
                return dos[4:6].reshape((2, self.nkpoints, self.nepoints))[
                    :, :, self.energy_window[0] : self.energy_window[1]
                ]
            return dos[3:5][:, self.energy_window[0] : self.energy_window[1]]

        elif self.case == 3:
            if self.k_resolved:
                if squeeze:
                    return (
                        np.sum(
                            dos[3:5].reshape((2, self.nkpoints, self.nepoints)), axis=1
                        )
                        / self.nkpoints
                    )[:, self.energy_window[0] : self.energy_window[1]]
                return dos[3:5].reshape((2, self.nkpoints, self.nepoints))[
                    :, :, self.energy_window[0] : self.energy_window[1]
                ]
            return dos[2:4][:, self.energy_window[0] : self.energy_window[1]]

        if self.k_resolved:
            if squeeze:
                return (
                    np.sum(dos[3].reshape((self.nkpoints, self.nepoints)), axis=0)
                    / self.nkpoints
                )[self.energy_window[0] : self.energy_window[1]]
            return dos[3].reshape((self.nkpoints, self.nepoints))[
                :, self.energy_window[0] : self.energy_window[1]
            ]
        return dos[2][self.energy_window[0] : self.energy_window[1]]

    @property
    def atoms(self):
        r"""
        List of atom's types.

        Returns
        -------
        atoms : list
            List of atom's types labels

            .. code-block:: python

                atoms = [atom1, atom2, ...]
        """

        return [atom for atom in self._atoms]

    def atom_numbers(self, atom: str):
        r"""
        List of atom's number for particular atom type.

        Parameters
        ----------
        atom : str
            Label of atom.

        Returns
        -------
        numbers : list

            .. code-block:: python

                numbers = [n_1, n_2, n_3, ...]
        """

        return self._atoms[atom]

    def wfcs(self, atom: str, atom_number: int = None):
        r"""
        Return list of wave function labels for particular atom.

        Parameters
        ----------
        atom : str
            Label of an atom.
        atom_number : int, default None
            Number of an atom. If ``None`` then return wfc and wfc numbers of first atom.

        Returns
        -------
        wfcs : list
            List of wave function labels and numbers

            .. code-block:: python

                wfcs = [(wfc_label, wfc_number), ...]

        """
        if atom_number is None:
            atom_number = self.atom_numbers(atom)[0]
        return self._wfcs[(atom, atom_number)]

    def pdos(self, atom, wfc, wfc_number, atom_numbers=None) -> PDOSQE:
        r"""
        Projected density of states for a particular atom.

        Parameters
        ----------
        atom : str
            Name of the atom type.
        wfc : str
            Name of the projector wave function.
        wfc_numbers : int
            Number of wave function projector.
        atom_numbers : list or int, default None
            If ``None``, then PDOS summed over all atom numbers for ``atom``.

        Returns
        -------
        pdos : :py:class:`PDOSQE`
            Partial density of states, orbital-resolved.
        """

        if atom_numbers is None:
            atom_numbers = self.atom_numbers(atom)
        elif isinstance(atom_numbers, int):
            if atom_numbers in self.atom_numbers(atom):
                atom_numbers = [atom_numbers]
            else:
                raise ValueError(f"There is no {atom}{atom_numbers} in PDOS.")

        for i, atom_number in enumerate(atom_numbers):
            path = join(
                self._input_path,
                f"{self.seedname}.pdos_atm#{atom_number}({atom})_wfc#{wfc_number}({wfc})",
            )
            if i == 0:
                pdos = np.loadtxt(path, skiprows=1).T
            else:
                pdos += np.loadtxt(path, skiprows=1).T
        if self.case in [2, 3]:
            if self.k_resolved:
                ldos = pdos[2:4].reshape(2, self.nkpoints, self.nepoints)[
                    :, :, self.energy_window[0] : self.energy_window[1]
                ]
                pdos = pdos[4:].reshape(
                    (pdos.shape[0] - 4) // 2, 2, self.nkpoints, self.nepoints
                )[:, :, :, self.energy_window[0] : self.energy_window[1]]
            else:
                ldos = pdos[1:3][:, self.energy_window[0] : self.energy_window[1]]
                pdos = pdos[3:].reshape((pdos.shape[0] - 3) // 2, 2, self.nepoints)[
                    :, :, self.energy_window[0] : self.energy_window[1]
                ]
        else:
            if self.k_resolved:
                ldos = pdos[2].reshape(self.nkpoints, self.nepoints)[
                    :, self.energy_window[0] : self.energy_window[1]
                ]
                pdos = pdos[3:].reshape(
                    pdos.shape[0] - 3, self.nkpoints, self.nepoints
                )[:, :, self.energy_window[0] : self.energy_window[1]]
            else:
                ldos = pdos[1][self.energy_window[0] : self.energy_window[1]]
                pdos = pdos[2:][:, self.energy_window[0] : self.energy_window[1]]

        return PDOSQE(
            energy=self.energy,
            pdos=pdos,
            projectors_group=wfc,
            ldos=ldos,
            spin_pol=self.case in [2, 3],
        )

    def plot_pdos_tot(
        self,
        output_name,
        interactive=False,
        efermi=0,
        xlim=None,
        ylim=None,
        save_pickle=False,
    ):
        r"""
        Plot total DOS vs total PDOS.

        Parameters
        ----------
        output_name : str
            Name of the output file.
        interactive : bool, default False
            Whether to plot in interactive matplotlib window.
        efermi : float, default 0
            Fermi energy. Zero of energy scale is shifted to it.
        xlim : tuple, default None
            Limits for energy scale.
        ylim : tuple, default None
            Limits for dos scale.
        save_pickle : bool, default False
            Whether to save figure as a .pickle file.
            Helps for custom modification of particular figures.
        """
        fig, ax = plt.subplots(figsize=(8, 4))

        ax.set_xlabel("Energy, eV", fontsize=18)
        ax.set_ylabel("DOS, states/eV", fontsize=18)
        if xlim is None:
            xlim = (np.amin(self.energy), np.amax(self.energy))
        ax.set_xlim(*tuple(xlim))
        if ylim is not None:
            ax.set_ylim(*tuple(ylim))
        if efermi != 0:
            ax.set_title(f"DOS vs PDOS (0 is Fermi energy)", fontsize=18)
        else:
            ax.set_title(f"DOS vs PDOS (0 is 0)", fontsize=18)
        ax.vlines(
            0,
            0,
            1,
            transform=ax.get_xaxis_transform(),
            color="grey",
            linewidths=0.5,
            linestyles="dashed",
        )

        if self.case in [1, 4]:
            ax.fill_between(
                self.energy,
                0,
                self.total_dos(squeeze=True),
                lw=0,
                color="grey",
                alpha=0.3,
                label="DOS",
            )
            ax.plot(
                self.energy,
                self.total_pdos(squeeze=True),
                "-",
                lw=1,
                color="black",
                alpha=0.7,
                label="PDOS",
            )
            ncol = 1
        if self.case == 2:
            ax.fill_between(
                self.energy,
                0,
                self.total_dos(squeeze=True)[0],
                lw=0,
                color="blue",
                alpha=0.3,
                label="DOS (up)",
            )
            ax.fill_between(
                self.energy,
                0,
                -self.total_dos(squeeze=True)[1],
                lw=0,
                color="red",
                alpha=0.3,
                label="DOS (down)",
            )

            ax.plot(
                self.energy,
                self.total_pdos(squeeze=True)[0],
                "-",
                lw=1,
                color="blue",
                alpha=0.7,
                label="PDOS (up)",
            )
            ax.plot(
                self.energy,
                -self.total_pdos(squeeze=True)[1],
                "-",
                lw=1,
                color="red",
                alpha=0.7,
                label="PDOS (down)",
            )
            ncol = 2
        if self.case == 3:
            ax.fill_between(
                self.energy,
                0,
                self.total_dos(squeeze=True),
                lw=0,
                color="grey",
                alpha=0.3,
                label="DOS",
            )
            ax.fill_between(
                self.energy,
                -self.total_dos(squeeze=True),
                0,
                lw=0,
                color="grey",
                alpha=0.3,
                label="-DOS",
            )

            ax.plot(
                self.energy,
                self.total_pdos(squeeze=True)[0],
                "-",
                lw=1,
                color="blue",
                alpha=0.7,
                label="PDOS (up)",
            )
            ax.plot(
                self.energy,
                -self.total_pdos(squeeze=True)[1],
                "-",
                lw=1,
                color="red",
                alpha=0.7,
                label="PDOS (down)",
            )
            ncol = 1

        if interactive:
            ax.legend(loc="best", ncol=ncol, draggable=True)
            plt.show()
        else:
            ax.legend(loc="best", ncol=ncol)
            png_path = f"{output_name}.png"
            plt.savefig(png_path, dpi=600, bbox_inches="tight")
            if save_pickle:
                import pickle

                with open(f"{png_path}.pickle", "wb") as file:
                    pickle.dump(fig, file)
        plt.close()


class DOSIterator:
    def __init__(self, dos: DOSQE) -> None:
        self.list = []
        for atom in dos.atoms:
            for atom_number in dos.atom_numbers(atom):
                for wfc, wfc_number in dos.wfcs(atom, atom_number):
                    self.list.append([atom, atom_number, wfc, wfc_number])
        self._index = 0

    def __next__(self) -> str:
        if self._index < len(self.list):
            result = self.list[self._index]
            self._index += 1
            return result
        raise StopIteration

    def __iter__(self):
        return self



from math import sqrt

import numpy as np

from rad_tools.kpoints.high_symmetry_point import HighSymmetryPoints


class KPoints(HighSymmetryPoints):
    r"""
    K-points
    """

    def __init__(self, kpoints=None, path=None) -> None:
        super().__init__(kpoints, path)

    @property
    def labels(self):
        r"""
        Labels of the high symmetry k points, ready to plot.

        Returns
        -------
        label_list : list of str
            List of high symmetry k points labels.
        """
        label_list = []
        for i in range(0, len(self.path)):
            if i != 0:
                label_list[-1] += self._PLOT_LITERALS[self.path[i][0]]
            else:
                label_list.append(self._PLOT_LITERALS[self.path[i][0]])
            for j in range(1, len(self.path[i])):
                label_list.append(self._PLOT_LITERALS[self.path[i][j]])
            if i != len(self.path) - 1:
                label_list[-1] += "$\\vert$"
        return label_list

    def labels_coordinates(self, b1=None, b2=None, b3=None):
        r"""
        Coordinates of the high symmetry k points.

        Returns
        -------
        coordinate_list : list of float
            List of high symmetry k points coordinates in reciprocal space
            if ``b1``, ``b2``, ``b3`` are specified, in relative otherwise.
        """

        coordinates_list = self.coordinates(n=1, b1=b1, b2=b2, b3=b3, linear=True)
        return np.concatenate(
            ([coordinates_list[0]], coordinates_list[1:-1:2], [coordinates_list[-1]])
        )

    def coordinates(self, n=100, b1=None, b2=None, b3=None, linear=False):
        r"""
        Coordinates of all k points in the path.

        Return coordinates in reciprocal space units if ``b1``, ``b2``, ``b3``
        are specified, otherwise return in relative to the reciprocal
        lattice vectors coordinates.

        Parameters
        ----------
        n : int
            Number of k points between two high symmetry points + 1.
        b1 : 1 x 3 array
            First reciprocal lattice vector.
        b2 : 1 x 3 array
            Second reciprocal lattice vector.
        b3 : 1 x 3 array
            Third reciprocal lattice vector.
        linear : bool, default False
            Return linear coordinates for the plot.

        Returns
        -------
        k_points : N x 3 array
            Array of N total k points in the path.

            :math:`N = i \cdot n`,
            where i - number of intervals between high symmetry k points.
        """

        point_list = []
        for i in range(0, len(self.path)):
            for j in range(1, len(self.path[i])):
                first_point = self.kpoints[self.path[i][j - 1]]
                diff = self.kpoints[self.path[i][j]] - self.kpoints[self.path[i][j - 1]]

                # Move from relative to absolute coordinates
                if b1 is not None and b2 is not None and b3 is not None:
                    first_point = (
                        first_point[0] * b1 + first_point[1] * b2 + first_point[2] * b3
                    )
                    diff = diff[0] * b1 + diff[1] * b2 + diff[2] * b3
                # Move from (kx, ky, kz) to |k|
                if linear:
                    diff = sqrt(np.sum(np.array(diff) ** 2))
                    if i == 0 and j == 1:
                        first_point = 0
                    else:
                        first_point = point_list[-1]

                for step in range(0, n + 1):
                    point_list.append(first_point + step / n * diff)

        return np.array(point_list)



#! /usr/local/bin/python3

from argparse import ArgumentParser
from os.path import join, split

import numpy as np

from rad_tools.routines import cprint


def manager(filename, span, output_path, output_name, no_colour=False):
    head, tail = split(filename)
    if output_path is None:
        output_path = head
    if output_name is None:
        output_name = tail + "_identified"
    separation_tolerance = 10e-8

    # Read atoms and centres
    atom_counter = {}
    with open(filename, "r") as file:
        n = int(file.readline())
        file_stats = file.readline()
        centres = []
        atoms = []
        for _ in range(0, n):
            line = file.readline().split()
            if line[0] == "X":
                centres.append([f"X", np.array(list(map(float, line[1:])))])
            else:
                if line[0] in atom_counter:
                    atom_counter[line[0]] += 1
                else:
                    atom_counter[line[0]] = 1
                atoms.append(
                    [
                        f"{line[0]}{atom_counter[line[0]]}",
                        np.array(list(map(float, line[1:]))),
                    ]
                )

    # Identify centres localization
    for centre in centres:
        min_span = 10000
        name = "None"
        for atom, a_coord in atoms:
            if np.linalg.norm((centre[1] - a_coord)) < min_span:
                min_span = np.linalg.norm((centre[1] - a_coord))
                name = atom
        if min_span - span > separation_tolerance:
            if no_colour:
                print(f"Centre {centre} unidentified, " + "try to increase span")
            else:
                cprint(
                    f"Centre {centre} unidentified, " + "try to increase span",
                    colour="yellow",
                )
            print(
                f"    span limit = {span}\n"
                + f"    minimum distance to the atom ({name}) = {min_span:.8f}\n"
            )
            centre[0] = f"None, closest: {name} ({min_span:.8f})"
        else:
            centre[0] = name

    # Write the output
    with open(join(output_path, output_name), "w") as file:
        file.write(f"{len(atoms) + len(centres):6.0f}\n")
        file.write(f"{file_stats}")
        for centre, coordinate in centres:
            file.write(
                f"X      "
                + f"{coordinate[0]:14.8f}   "
                + f"{coordinate[1]:14.8f}   "
                + f"{coordinate[2]:14.8f}"
                + f"   ->   {centre}\n"
            )
        for atom, coordinate in atoms:
            file.write(
                f"{atom.translate(str.maketrans('','','0123456789')):2}     "
                + f"{coordinate[0]:14.8f}   "
                + f"{coordinate[1]:14.8f}   "
                + f"{coordinate[2]:14.8f}"
                + f"   ->   {atom}\n"
            )


def create_parser():
    parser = ArgumentParser()
    parser.add_argument(
        "filename",
        type=str,
        help="Relative or absolute path to the _centres.xyz file",
    )
    parser.add_argument(
        "-s",
        "--span",
        type=float,
        default=0.1,
        help="Distance tolerance between centre and atom. (in Angstroms)",
    )
    parser.add_argument(
        "-op",
        "--output-path",
        type=str,
        default=None,
        help="Relative or absolute path to the folder for saving outputs.",
    )
    parser.add_argument(
        "-on",
        "--output-name",
        type=str,
        default=None,
        help="Seedname for the output files.",
    )
    parser.add_argument(
        "-nc",
        "--no-colour",
        action="store_true",
        default=False,
        help="Turn off coloured output.",
    )
    return parser



r"""
Exchange bond.
"""

from math import sqrt

import numpy as np


class Bond:
    r"""
    Exchange bond.

    If ``matrix`` is specified then ``iso``, ``aniso`` and ``dmi`` are
    ignored and derived from ``matrix``. If ``matrix`` is not specified
    then it is derived from ``iso``, ``aniso`` and ``dmi``.
    If nothing is provided exchange matrix is set to zero matrix.

    Parameters
    ----------
    iso : int or float, default None
        Value of isotropic exchange parameter.
    aniso : 3 x 3 array, None
        3 x 3 matrix of symmetric anisotropic exchange.
    dmi : 1 x 3 array, None
        Dzyaroshinsky-Moria interaction vector :math:`(D_x, D_y, D_z)`.
    matrix : 3 x 3 array, None
        Exchange matrix.

    Attributes
    ----------
    matrix : 3 x 3 array of floats
        Exchange matrix.

        .. code-block:: python

            [[J_xx, J_xy, J_xz],
             [J_yx, J_yy, J_yz],
             [J_zx, J_zy, J_zz]]

    symm_matrix : 3 x 3 array of floats
    asymm_matrix : 3 x 3 array of floats
    iso : float
    iso_matrix : 3 x 3 array of floats
    aniso : 3 x 3 array of floats
    aniso_diagonal : 1 x 3 array of floats
    aniso_diagonal_matrix : 3 x 3 array of floats
    dmi : 1 x 3 array of floats
    dmi_matrix : 3 x 3 array of floats
    dmi_module : float
    """

    def __init__(self, matrix=None, iso=None, aniso=None, dmi=None) -> None:
        self._matrix = np.zeros((3, 3), dtype=float)
        self._iso = 0.0
        self._aniso = np.zeros((3, 3), dtype=float)
        self._dmi = np.zeros(3, dtype=float)

        if matrix is None:
            self.iso = iso
            self.dmi = dmi
            self.aniso = aniso
        else:
            self.matrix = matrix

    @property
    def matrix(self):
        return self._matrix

    @matrix.setter
    def matrix(self, new_matrix):
        if new_matrix is not None:
            new_matrix = np.array(new_matrix, dtype=float)
            if new_matrix.shape != (3, 3):
                raise ValueError("Matrix shape has to be equal to (3, 3)")
        else:
            new_matrix = np.zeros((3, 3), dtype=float)
        self._matrix = new_matrix

    @property
    def symm_matrix(self):
        r"""
        Symmetric part of exchange matrix.

        .. math::

            J_symm = \dfrac{J + J^T}{2}
        """

        return (self.matrix + self.matrix.T) / 2

    @property
    def asymm_matrix(self):
        """
        Asymmetric part of exchange matrix.

        .. math::

            J_asymm = \dfrac{J - J^T}{2}
        """

        return (self.matrix - self.matrix.T) / 2

    @property
    def iso(self):
        r"""
        Value of isotropic exchange parameter.

        Derived from the exchange matrix (:math:`\mathbf{J}`) as

        .. math::
            J_{iso} = \dfrac{1}{3}Tr(\mathbf{J})
        """

        return np.trace(self.symm_matrix) / 3

    @iso.setter
    def iso(self, new_iso):
        if new_iso is not None:
            # correction is correct (see matrix update)
            new_iso = new_iso - self.iso
        else:
            new_iso = -self.iso
        self.matrix = self.matrix + (new_iso) * np.identity(3, dtype=float)

    @property
    def iso_matrix(self):
        r"""
        Isotropic part of the exchange matrix.

        Matrix form:

        .. code-block:: python

            [[J, 0, 0],
             [0, J, 0],
             [0, 0, J]]
        """

        return self.iso * np.identity(3, dtype=float)

    @property
    def aniso(self):
        r"""
        3 x 3 matrix of symmetric anisotropic exchange.

        .. code-block:: python

            [[J_xx, J_xy, J_xz],
             [J_xy, J_yy, J_yz],
             [J_xz, J_yz, J_zz]]

        Derived from the exchange matrix (:math:`\mathbf{J}`) as

        .. math::
            \mathbf{J}_{aniso} = \mathbf{J}_{symm} - \dfrac{1}{3}Tr(\mathbf{J})
            \cdot \mathbf{I}

        where :math:`\mathbf{I}` is a :math:`3\times3` identity matrix.
        """

        return self.symm_matrix - self.iso * np.identity(3, dtype=float)

    @aniso.setter
    def aniso(self, new_aniso):
        if new_aniso is not None:
            new_aniso = np.array(new_aniso, dtype=float)
            if new_aniso.shape != (3, 3):
                raise ValueError("Aniso matrix shape have " + "to be equal to (3, 3)")
            # correction is correct (see matrix update)
            new_aniso = new_aniso - self.aniso
        else:
            new_aniso = -self.aniso
        self.matrix = self.matrix + new_aniso

    @property
    def aniso_diagonal(self):
        r"""
        Diagonal part of the symmetric anisotropic exchange.

        .. code-block:: python

            [J_xx, J_yy, J_zz]
        """

        return np.diag(self.aniso)

    @property
    def aniso_diagonal_matrix(self):
        r"""
        Diagonal part of the symmetric anisotropic exchange matrix.

        .. code-block:: python

            [[J_xx, 0, 0],
             [0, J_yy, 0],
             [0, 0, J_zz]]
        """

        return np.diag(np.diag(self.aniso))

    @property
    def dmi(self):
        r"""
        Dzyaroshinsky-Moria interaction vector (Dx, Dy, Dz).

        .. code-block:: python

            [D_x, D_y, D_z]
        """

        return np.array(
            [self.asymm_matrix[1][2], self.asymm_matrix[2][0], self.asymm_matrix[0][1]],
            dtype=float,
        )

    @dmi.setter
    def dmi(self, new_dmi):
        if new_dmi is not None:
            new_dmi = np.array(new_dmi, dtype=float)
            if new_dmi.shape != (3,):
                raise ValueError(f"DMI have to be a 3 component vector. {new_dmi}")
            new_dmi = new_dmi - self.dmi
        else:
            new_dmi = -self.dmi
        dmi_matrix = np.array(
            [
                [0, new_dmi[2], -new_dmi[1]],
                [-new_dmi[2], 0, new_dmi[0]],
                [new_dmi[1], -new_dmi[0], 0],
            ],
            dtype=float,
        )
        self.matrix = self.matrix + dmi_matrix

    @property
    def dmi_matrix(self):
        r"""
        Asymmetric part of the exchange matrix.

        .. code-block:: python

            [[0, D_z, -D_y],
             [-D_z, 0, D_x],
             [D_y, -D_x, 0]]
        """

        return self.asymm_matrix

    @property
    def dmi_module(self):
        r"""
        Length of the DMI vector in the units of exchange interaction.
        """

        return sqrt(np.sum(self.dmi**2))

    def round(self, decimals=4):
        r"""Round exchange values.

        Parameters
        ----------
        accuracy : int, default 4
            number of decimals after the point.
        """

        self.matrix = np.round(self.matrix, decimals=decimals)

    def __add__(self, other):
        if isinstance(other, Bond):
            matrix = self.matrix + other.matrix
            return Bond(matrix=matrix)
        else:
            raise TypeError

    def __sub__(self, other):
        if isinstance(other, Bond):
            matrix = self.matrix - other.matrix
            return Bond(matrix=matrix)
        else:
            raise TypeError

    def __mul__(self, number):
        matrix = self.matrix * number
        return Bond(matrix=matrix)

    def __rmul__(self, number):
        return self.__mul__(number)

    def __truediv__(self, number):
        matrix = self.matrix / number
        return Bond(matrix=matrix)



#! /usr/local/bin/python3

import sys
from time import sleep
from rad_tools.routines import cprint
from rad_tools.score.rad_plot_dos_core import create_parser, manager

if __name__ == "__main__":
    parser = create_parser()
    args = parser.parse_args()
    manager(**vars(args))
    if sys.platform == "win32":
        cprint("Press Enter to continue", colour="green")
        input()



from argparse import ArgumentParser
from calendar import month_name
from datetime import datetime
from os import makedirs
from os.path import abspath, join

from rad_tools import __version__ as version
from rad_tools.io import read_tb2j_model
from rad_tools.io.internal import read_template
from rad_tools.routines import OK, RESET


def manager(
    input_filename,
    template_file,
    output_path=".",
    output_name=None,
    decimals=4,
    force_symmetry=False,
    isotropic=False,
    anisotropic=False,
    matrix=False,
    dmi=False,
    all=False,
    verbose=False,
):
    r"""
    Main function call of the tb2j-extractor.py script.

    Full documentation on the behaviour is available
    :ref:`here <tb2j-extractor>`. Parameters of the function directly
    corresponds to the arguments of the script.

    If you want to have the behaviour of the tb2j-extractor.py script
    but in a format of a function call use this function.
    """

    try:
        makedirs(output_path)
    except FileExistsError:
        pass

    cd = datetime.now()

    if all:
        isotropic = True
        anisotropic = True
        matrix = True
        dmi = True

    model = read_tb2j_model(input_filename, quiet=not verbose)
    template = read_template(template_file)
    summary_txt = model.summary_as_txt(
        template=template,
        decimals=decimals,
        force_symmetry=force_symmetry,
        isotropic=isotropic,
        anisotropic=anisotropic,
        out_matrix=matrix,
        out_dmi=dmi,
    )

    if output_name is not None:
        with open(join(output_path, output_name + ".txt"), "w") as out_file:
            out_file.write(
                f"Exchange values are extracted from: {input_filename}\n"
                + f"on {cd.day} {month_name[cd.month]} {cd.year}"
                + f" at {cd.hour}:{cd.minute}:{cd.second} by rad-tools {version}\n\n"
            )
            out_file.write(summary_txt)
        print(
            f"{OK}Extracted exchange info is in "
            + f"{abspath(join(output_path, output_name + '.txt'))}{RESET}"
        )
    else:
        print(f"{summary_txt}")


def create_parser():
    parser = ArgumentParser(
        description="Script for extracting of template-based model from TB2J results."
    )

    parser.add_argument(
        "-if",
        "--input-filename",
        metavar="filename",
        type=str,
        required=True,
        help="Relative or absolute path to the 'exchange.out' file, "
        + "including the name and extension of the file itself.",
    )
    parser.add_argument(
        "-tf",
        "--template-file",
        metavar="filename",
        type=str,
        required=True,
        help="Relative or absolute path to the template file, "
        + "including the name and extension of the file.",
    )
    parser.add_argument(
        "-op",
        "--output-path",
        metavar="path",
        type=str,
        default=".",
        help="Relative or absolute path to the folder for saving outputs.",
    )
    parser.add_argument(
        "-on",
        "--output-name",
        metavar="filename",
        type=str,
        default=None,
        help="Seedname for the output files.",
    )
    parser.add_argument(
        "-d",
        "--decimals",
        metavar="n",
        type=int,
        default=4,
        help="Decimals after the comma for the exchange values.",
    )
    parser.add_argument(
        "-fs",
        "--force-symmetry",
        action="store_true",
        default=False,
        help="Whenever to force the symmetry of the template on the model.",
    )
    parser.add_argument(
        "-i",
        "--isotropic",
        action="store_true",
        default=False,
        help="Whenever to output isotropic exchange.",
    )
    parser.add_argument(
        "-a",
        "--anisotropic",
        action="store_true",
        default=False,
        help="Whenever to output anisotropic exchange.",
    )
    parser.add_argument(
        "-m",
        "--matrix",
        action="store_true",
        default=False,
        help="Whenever to output whole matrix exchange.",
    )
    parser.add_argument(
        "-dmi",
        action="store_true",
        default=False,
        help="Whenever to output DMI exchange.",
    )
    parser.add_argument(
        "-all",
        action="store_true",
        default=False,
        help="Whenever to all types of exchange.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        default=False,
        help="Verbose output, propagates to the called methods.",
    )

    return parser



from setuptools import find_packages, setup

from rad_tools import __version__

__version__ = __version__

with open("README.rst", "r") as file:
    long_description = ""
    for line in file:
        long_description += line

setup(
    name="rad-tools",
    version=__version__,
    description="Collection of scripts from my PhD",
    long_description=long_description,
    author="Andrey Rybakov",
    author_email="rybakov.ad@icloud.com",
    license="MIT license",
    url="https://github.com/adrybakov/rad-tools",
    download_url="https://github.com/adrybakov/rad-tools.git",
    packages=find_packages(),
    scripts=[
        "scripts/tb2j-plotter.py",
        "scripts/phonopy-plotter.py",
        "scripts/tb2j-extractor.py",
        "scripts/rad-make-template.py",
        "scripts/rad-plot-dos.py",
        "scripts/identify-wannier-centres.py",
        "scripts/compute-energies.py",
    ],
    install_requires=["numpy", "matplotlib", "tqdm"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Chemistry",
        "Topic :: Scientific/Engineering :: Physics",
        "License :: OSI Approved :: MIT License",
    ],
    python_requires=">=3.6",
)



import re
from argparse import ArgumentParser
from os import makedirs, walk
from os.path import abspath, join

from tqdm import tqdm

from rad_tools.dos import DOSQE, plot_projected, PDOS
from rad_tools.routines import cprint


def detect_seednames(input_path):
    r"""
    Analyze input folder, detects seednames for the dos output files.

    Parameters
    ----------
    input_path : str
        Directory with DOS files.

    Returns
    -------
    seednames : list
        List of seednames found in ``input_path``.
    """

    pattern = ".pdos_atm#[0-9]*\\([a-zA-Z]*\\)_wfc#[0-9]*\\([spdf_0-9j.]*\\)"

    # Get list of files in the folder
    files = []
    for dirpath, dirnames, filenames in walk(input_path):
        files.extend(filenames)
        break

    seednames = set()
    for file in files:
        if ".pdos_tot" in file and ".pdos_tot" == file[-9:]:
            seednames.add(file[:-9])
        elif re.match(f".*{pattern}$", file):
            seednames.add(re.split(f"{pattern}$", file)[0])
    seednames = list(seednames)

    return seednames


def manager(
    input_path,
    seedname=None,
    output_path=".",
    energy_window=None,
    dos_window=None,
    efermi=0.0,
    separate=False,
    relative=False,
    normalize=False,
    verbose=False,
    interactive=False,
    save_pickle=False,
    save_txt=False,
):
    makedirs(output_path, exist_ok=True)

    suffix = ""
    if relative:
        suffix += "-relative"
    if normalize:
        suffix += "-normalized"
    if separate:
        suffix += "-separate"

    # Detect seednames if not provided.
    if seedname is None:
        seednames = detect_seednames(input_path)
        print(f"Following DOS seednames (filpdos) are detected:")
        for item in seednames:
            cprint(f"   * {item}", colour="yellow")
    else:
        seednames = [seedname]

    # Work with each seedname.
    for s_i, seedname in enumerate(seednames):
        cprint(
            f"({s_i + 1}/{len(seednames)}) Start to work with {seedname} seedname",
            colour="yellow",
        )
        # Preparations
        output_root = join(output_path, f"{seedname}{suffix}")
        makedirs(output_root, exist_ok=True)

        # Load DOS data.
        dos = DOSQE(seedname, input_path, energy_window=energy_window)
        print(f"{dos.casename} case detected.")
        for atom in dos.atoms:
            print(f"    {len(dos.atom_numbers(atom))} of {atom} detected")

        # Plot PDOS vs DOS
        cprint("Total DOS vs total PDOS", colour="yellow")
        dos.plot_pdos_tot(
            output_name=join(output_root, "pdos-vs-dos"),
            interactive=interactive,
            efermi=efermi,
            xlim=energy_window,
            ylim=dos_window,
            save_pickle=save_pickle,
        )
        print(f"  Result is in {join(output_root, 'pdos-vs-dos')}")

        # Plot PDOS for each atom/wfc
        cprint("Orbital-resolved PDOS:", colour="yellow")
        local_output = join(output_root, "orbital-resolved")
        makedirs(local_output, exist_ok=True)
        data = {}
        for atom, atom_number, wfc, wfc_number in dos:
            if atom not in data:
                data[atom] = []
            data[atom].append((wfc, wfc_number))

        # Avoid repetitions
        for atom in data:
            data[atom] = list(set(data[atom]))

        for atom in data:
            for wfc, wfc_number in data[atom]:
                if separate:
                    atom_numbers = dos.atom_numbers(atom)
                else:
                    atom_numbers = [None]

                for atom_number in tqdm(
                    atom_numbers, desc=f"  {atom} {wfc} #{wfc_number}"
                ):
                    if separate:
                        atom_name = f"{atom}#{atom_number}"
                    else:
                        atom_name = atom

                    if efermi == 0:
                        title = f"PDOS for {atom_name} ({wfc} #{wfc_number}) (0 is 0)"
                    else:
                        title = f"PDOS for {atom_name} ({wfc} #{wfc_number}) (0 is Fermi energy)"

                    pdos = dos.pdos(
                        atom=atom,
                        wfc=wfc,
                        wfc_number=wfc_number,
                        atom_numbers=atom_number,
                    )
                    if save_txt:
                        pdos.dump_txt(
                            join(local_output, f"{atom_name}_{wfc}#{wfc_number}.txt")
                        )
                    plot_projected(
                        pdos=pdos,
                        efermi=efermi,
                        output_name=join(
                            local_output, f"{atom_name}_{wfc}#{wfc_number}"
                        ),
                        title=title,
                        xlim=energy_window,
                        ylim=dos_window,
                        relative=relative,
                        normalize=normalize,
                        interactive=interactive,
                        save_pickle=save_pickle,
                    )
        print(f"  Results are in {abspath(local_output)}")

        # Plot wfc contribution into each atom
        cprint("Orbital's contribution for each atom.", colour="yellow")
        local_output = join(output_root, "atom-resolved")
        makedirs(local_output, exist_ok=True)

        for atom in dos.atoms:
            if separate:
                atom_numbers = dos.atom_numbers(atom)
            else:
                atom_numbers = [None]
            for atom_number in tqdm(atom_numbers, desc=f"  {atom}"):
                if separate:
                    atom_name = f"{atom}#{atom_number}"
                else:
                    atom_name = atom
                projectors = []
                pdos = []
                for wfc, wfc_number in dos.wfcs(atom, atom_number):
                    projectors.append(f"{wfc} #{wfc_number}")
                    pdos.append(dos.pdos(atom, wfc, wfc_number, atom_number).ldos)

                pdos = PDOS(
                    energy=dos.energy,
                    pdos=pdos,
                    projectors_group=atom_name,
                    projectors=projectors,
                    spin_pol=dos.case in [2, 3],
                )
                if efermi == 0:
                    title = f"PDOS for {atom_name} (0 is 0)"
                else:
                    title = f"PDOS for {atom_name} (0 is Fermi energy)"
                if save_txt:
                    pdos.dump_txt(join(local_output, f"{atom_name}.txt"))
                plot_projected(
                    pdos=pdos,
                    efermi=efermi,
                    output_name=join(local_output, atom_name),
                    title=title,
                    xlim=energy_window,
                    ylim=dos_window,
                    relative=relative,
                    normalize=normalize,
                    interactive=interactive,
                    save_pickle=save_pickle,
                )
        print(f"  Results are in {abspath(local_output)}")

        # Plot atom contributions into total PDOS
        cprint("Atom's contributions into total PDOS:", colour="yellow")
        projectors = []
        pdos = []
        for atom in dos.atoms:
            if separate:
                atom_numbers = dos.atom_numbers(atom)
            else:
                atom_numbers = [None]
            for atom_number in atom_numbers:
                if separate:
                    atom_name = f"{atom}#{atom_number}"
                else:
                    atom_name = atom
                projectors.append(atom_name)
                for i, (wfc, wfc_number) in enumerate(dos.wfcs(atom, atom_number)):
                    if i == 0:
                        ldos = dos.pdos(atom, wfc, wfc_number, atom_number).ldos
                    else:
                        ldos += dos.pdos(atom, wfc, wfc_number, atom_number).ldos
                pdos.append(ldos)

        pdos = PDOS(
            energy=dos.energy,
            pdos=pdos,
            projectors_group="Total PDOS",
            projectors=projectors,
            spin_pol=dos.case in [2, 3],
        )
        if efermi == 0:
            title = f"Atom contribution in PDOS (0 is 0)"
        else:
            title = f"Atom contribution in PDOS (0 is Fermi energy)"
        if save_txt:
            pdos.dump_txt(join(output_root, "atomic-contributions.txt"))
        plot_projected(
            pdos=pdos,
            efermi=efermi,
            output_name=join(output_root, "atomic-contributions"),
            title=title,
            xlim=energy_window,
            ylim=dos_window,
            relative=relative,
            normalize=normalize,
            interactive=interactive,
            save_pickle=save_pickle,
        )
        print(f"  Result is in {abspath(local_output)}")


def create_parser():
    parser = ArgumentParser(
        description="Script for visualisation of density of states."
    )
    parser.add_argument(
        "-ip",
        "--input-path",
        metavar="path",
        type=str,
        required=True,
        help="Relative or absulute path to the folder with dos files.",
    )
    parser.add_argument(
        "-s",
        "--seedname",
        metavar="name",
        type=str,
        default=None,
        help="Prefix for output files containing PDOS(E). "
        + "As specified in the QE projwfc.x input file.",
    )
    parser.add_argument(
        "-op",
        "--output-path",
        metavar="path",
        type=str,
        default=".",
        help="Relative or absolute path to the folder for saving outputs.",
    )
    parser.add_argument(
        "-ew",
        "--energy-window",
        metavar=("min", "max"),
        type=float,
        nargs=2,
        default=None,
        help="Energy window for the plots. "
        + "By default whole range present in the files is plotted.",
    )
    parser.add_argument(
        "-dw",
        "--dos-window",
        metavar=("min", "max"),
        type=float,
        nargs=2,
        default=None,
        help="DOS window for the plots. "
        + "By default whole range present in the files is plotted.",
    )
    parser.add_argument(
        "-ef",
        "--efermi",
        metavar="energy",
        type=float,
        default=0.0,
        help="Fermi energy, zero by default.",
    )
    parser.add_argument(
        "-sep",
        "--separate",
        action="store_true",
        default=False,
        help="Whenever to plot projected DOS for each atom of the same type separately.",
    )
    parser.add_argument(
        "-r",
        "--relative",
        action="store_true",
        default=False,
        help="Whenever to use relative style.",
    )
    parser.add_argument(
        "-n",
        "--normalize",
        action="store_true",
        default=False,
        help="Whenever to use normalize relative style.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        default=False,
        help="Verbose output, propagates to the called methods.",
    )
    parser.add_argument(
        "-i",
        "--interactive",
        action="store_true",
        default=False,
        help="Interactive plotting.",
    )
    parser.add_argument(
        "-sp",
        "--save-pickle",
        action="store_true",
        default=False,
        help="Whenever to save figures as .pickle files.",
    )
    parser.add_argument(
        "-st",
        "--save-txt",
        action="store_true",
        default=False,
        help="Whenever to save some data as txt files.",
    )

    return parser



from argparse import ArgumentParser
from math import sqrt
from os import makedirs
from os.path import abspath, join

import numpy as np
from matplotlib import pyplot as plt

from rad_tools.io.internal import read_template
from rad_tools.io.tb2j import read_exchange_model
from rad_tools.routines import OK, RESET, atom_mark_to_latex, rot_angle


def manager(
    input_filename,
    output_path=".",
    output_name="exchange",
    what_to_plot="iso",
    draw_cells=False,
    min_distance=None,
    max_distance=None,
    distance=None,
    template_file=None,
    R_vector=None,
    scale_atoms=1,
    scale_data=1,
    title=None,
    force_symmetry=False,
    verbose=False,
):
    r"""
    Main function call of the tb2j-plotter.py script.

    Full documentation on the behaviour is available
    :ref:`here <tb2j-plotter>`. Parameters of the function directly
    corresponds to the arguments of the script.

    If you want to have the behaviour of the tb2j-plotter.py script
    but in a format of a function call use this function.
    """

    if force_symmetry and template_file is None:
        raise ValueError("--force-symmetry option requires a template file.")

    if distance is not None:
        min_distance = distance
        max_distance = distance

    try:
        makedirs(output_path)
    except FileExistsError:
        pass

    if R_vector is not None:
        R_vector = np.array(R_vector[: len(R_vector) // 3 * 3], dtype=int).reshape(
            (len(R_vector) // 3, 3)
        )
        R_vector = list(map(tuple, R_vector.tolist()))

    model = read_exchange_model(input_filename, quiet=not verbose)
    if template_file is not None:
        template = read_template(template_file)
    else:
        template = None

    if force_symmetry:
        model.force_symmetry(template=template)

    model.filter(
        min_distance=min_distance,
        max_distance=max_distance,
        R_vector=R_vector,
        template=template,
    )

    dummy = True
    ha = "center"
    x_min, y_min, z_min, x_max, y_max, z_max = model.space_dimensions
    X = max(abs(x_min), abs(x_max))
    Y = max(abs(y_min), abs(y_max))
    if X == 0 and Y == 0:
        X = Y = 1
    linewidth = 1
    fontsize = 11
    if what_to_plot == "all":
        wtps = ["iso", "distance", "dmi"]
    else:
        wtps = [what_to_plot]
    for wtp in wtps:
        if X < Y:
            fig = plt.figure(figsize=(6.4 * X / Y, 4.8))
        else:
            fig = plt.figure(figsize=(6.4, 4.8 * Y / X))
        ax = fig.add_axes([0.1, 0.1, 0.8, 0.8])
        ax.set_aspect("equal")
        ax.set_xlabel("x, Angstroms")
        ax.set_ylabel("y, Angstroms")

        for atom1, atom2, R in model:
            bond = model[(atom1, atom2, R)]
            dis = model.get_distance(atom1, atom2, R)
            x1, y1, z1 = model.get_atom_coordinates(atom1)
            x2, y2, z2 = model.get_atom_coordinates(atom2, R)
            xm = (x1 + x2) / 2
            ym = (y1 + y2) / 2
            zm = (z1 + z2) / 2

            ax.scatter(x1, y1, s=100 * fig.dpi / 72.0, c="white")
            ax.scatter(x2, y2, s=100 * fig.dpi / 72.0, c="white")

            ax.text(
                x2,
                y2,
                atom_mark_to_latex(atom2),
                va="center",
                ha="center",
                fontsize=1.5 * fontsize * scale_atoms,
            )
            ax.text(
                x1,
                y1,
                atom_mark_to_latex(atom1),
                va="center",
                ha="center",
                fontsize=1.5 * fontsize * scale_atoms,
                color="#A04F4D",
            )
            if wtp == "iso":
                ax.text(
                    xm,
                    ym,
                    str(round(bond.iso, 4)),
                    va="bottom",
                    ha=ha,
                    rotation_mode="anchor",
                    rotation=rot_angle(x2 - x1, y2 - y1, dummy=dummy),
                    fontsize=fontsize * scale_data,
                )
            elif wtp == "distance":
                ax.text(
                    xm,
                    ym,
                    str(round(dis, 4)),
                    va="bottom",
                    ha=ha,
                    rotation_mode="anchor",
                    rotation=rot_angle(x2 - x1, y2 - y1, dummy=dummy),
                    fontsize=fontsize * scale_data,
                )
            elif wtp == "dmi":
                ax.text(
                    xm,
                    ym,
                    str(round(sqrt(np.sum(bond.dmi**2)), 4)),
                    va="bottom",
                    ha=ha,
                    rotation_mode="anchor",
                    rotation=rot_angle(x2 - x1, y2 - y1, dummy=dummy),
                    fontsize=fontsize * scale_data,
                )

        xlims = (ax.get_xlim()[0] - 0.5, ax.get_xlim()[1] + 0.5)
        ylims = (ax.get_ylim()[0] - 0.5, ax.get_ylim()[1] + 0.5)

        if draw_cells:
            cells = model.cell_list
            a_x, a_y, a_z = tuple(model.cell[0])
            b_x, b_y, b_z = tuple(model.cell[1])
            Rx_min = 0
            Rx_max = 0
            Ry_min = 0
            Ry_max = 0
            Rz_min = 0
            Rz_max = 0
            for Rx, Ry, Rz in cells:
                Rx_min = min(Rx, Rx_min)
                Rx_max = max(Rx, Rx_max)
                Ry_min = min(Ry, Ry_min)
                Ry_max = max(Ry, Ry_max)
                Rz_min = min(Rz, Rz_min)
                Rz_max = max(Rz, Rz_max)

            for i in range(Rx_min, Rx_max + 2):
                for j in range(Ry_min, Ry_max + 2):
                    for k in range(Rz_min, Rz_max + 2):
                        ax.plot(
                            [Rx_min * a_x + j * b_x, (Rx_max + 1) * a_x + j * b_x],
                            [Rx_min * a_y + j * b_y, (Rx_max + 1) * a_y + j * b_y],
                            linewidth=linewidth,
                            color="#DFDFDF",
                        )
                        ax.plot(
                            [i * a_x + Ry_min * b_x, i * a_x + (Ry_max + 1) * b_x],
                            [i * a_y + Ry_min * b_y, i * a_y + (Ry_max + 1) * b_y],
                            linewidth=linewidth,
                            color="#DFDFDF",
                        )
            ax.plot(
                np.array([0, a_x, a_x + b_x, b_x, 0]),
                np.array([0, a_y, a_y + b_y, b_y, 0]),
                linewidth=linewidth,
                color="#CA7371",
            )

        ax.set_xlim(*xlims)
        ax.set_ylim(*ylims)

        if title is not None:
            ax.set_title(title, fontsize=1.5 * fontsize)

        png_path = join(output_path, f"{output_name}.{wtp}.png")
        plt.savefig(png_path, dpi=400, bbox_inches="tight")
        print(f"{OK}2D plot with {wtp} is in {abspath(png_path)}{RESET}")


def create_parser():
    parser = ArgumentParser(description="Script for visualization of TB2J results")

    parser.add_argument(
        "-if",
        "--input-filename",
        metavar="filename",
        type=str,
        required=True,
        help="Relative or absolute path to the 'exchange.out' file,"
        + "including the name and extension of the file itself.",
    )
    parser.add_argument(
        "-op",
        "--output-path",
        metavar="path",
        type=str,
        default=".",
        help="Relative or absolute path to the folder for saving outputs.",
    )
    parser.add_argument(
        "-on",
        "--output-name",
        metavar="filename",
        type=str,
        default="exchange",
        help="Seedname for the output files.",
    )
    parser.add_argument(
        "-wtp",
        "--what-to-plot",
        metavar="value",
        type=str,
        choices=["all", "iso", "distance", "dmi"],
        default="all",
        help="Type of data for display.",
    )
    parser.add_argument(
        "-dc",
        "--draw-cells",
        action="store_true",
        default=False,
        help="Whenever to draw the cells.",
    )
    parser.add_argument(
        "-R",
        "--R-vector",
        metavar="R1a R1b R1c",
        type=int,
        nargs="*",
        default=None,
        help="R vectors for filtering the model.",
    )
    parser.add_argument(
        "-maxd",
        "--max-distance",
        metavar="distance",
        type=float,
        default=None,
        help="(<=) Maximum distance.",
    )
    parser.add_argument(
        "-mind",
        "--min-distance",
        metavar="distance",
        type=float,
        default=None,
        help="(>=) Minimum distance.",
    )
    parser.add_argument(
        "-d",
        "--distance",
        metavar="distance",
        type=float,
        default=None,
        help="(=) Exact distance.",
    )
    parser.add_argument(
        "-tf",
        "--template-file",
        metavar="filename",
        type=str,
        default=None,
        help="Relative or absolute path to the template file, "
        + "including the name and extension of the file.",
    )
    parser.add_argument(
        "-sa",
        "--scale-atoms",
        metavar="factor",
        default=1,
        type=float,
        help="Scale for the size of atom marks.",
    )
    parser.add_argument(
        "-sd",
        "--scale-data",
        metavar="factor",
        default=1,
        type=float,
        help="Scale for the size of data text.",
    )
    parser.add_argument(
        "-t",
        "--title",
        metavar="title",
        default=None,
        type=str,
        help="Title for the plots.",
    )
    parser.add_argument(
        "-fs",
        "--force-symmetry",
        default=False,
        action="store_true",
        help="Force the Exchange model to have the symmetry of the template.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        default=False,
        help="Verbose output, propagates to the called methods.",
    )

    return parser
